import os
import re
import random
import json
import pickle
import warnings
import math
import tempfile
from pathlib import Path
from zipfile import ZipFile
import traceback
import arcgis
from arcgis.features import FeatureLayer
from .._utils.tabular_data import TabularDataObject, explain_prediction, add_h3

try:
    import sklearn
    from sklearn import *
    from sklearn.preprocessing import LabelEncoder
    import pandas as pd
    import warnings
    from .._fairlearn import _fairlearn
    from .._fairlearn import _reweigh
    from fairlearn.postprocessing import ThresholdOptimizer

    from fairlearn.reductions import (
        ExponentiatedGradient,
        DemographicParity,
        EqualizedOdds,
        BoundedGroupLoss,
        ZeroOneLoss,
        SquareLoss,
        GridSearch,
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        import xgboost
    import lightgbm
    import catboost

    HAS_ML_DEPS = True
except:
    missing_deps_trace = traceback.format_exc()
    HAS_ML_DEPS = False

HAS_FAST_PROGRESS = True
try:
    from fastprogress.fastprogress import progress_bar
except:
    HAS_FAST_PROGRESS = False

_PROTOCOL_LEVEL = 2
_FAIRNESS_ARGS_NOT_DICT = "Fairness args must be a dictionary"
_FAIRNESS_ARGS_KEY_NOT_FOUND = "Fairness args key not found"
_DEGENERATE_LABEL_FOR_SENSITIVE_FEATURE = "ValueError: The sensitive feature encountered a degenerate label. A degenerate label typically refers to a label or category within a dataset that has very little variation or diversity, making it less informative for machine learning or statistical analysis."


def _get_model_type(model_type):
    if not model_type:
        raise Exception("Invalid model type.")

    if not isinstance(model_type, str):
        return model_type

    if model_type.startswith("sklearn."):
        # raise Exception("Invalid model_type.")
        model_type = model_type.replace("sklearn.", "")
        module = model_type.split(".")[0]
        if len(model_type.split(".")) > 1:
            model = model_type.split(".")[1]
        else:
            raise Exception("Invalid model_type.")
        if not hasattr(sklearn, module) or not hasattr(getattr(sklearn, module), model):
            raise Exception("Invalid model_type.")

        model = getattr(getattr(sklearn, module), model)

    elif model_type.startswith("xgboost."):
        model_type = model_type.replace("xgboost.", "")
        if len(model_type.split(".")) > 0:
            model = model_type.split(".")[0]
        else:
            raise Exception("Invalid model_type.")
        if not hasattr(xgboost, model):
            raise Exception("Invalid model_type.")

        model = getattr(xgboost, model)

    elif model_type.startswith("lightgbm."):
        model_type = model_type.replace("lightgbm.", "")
        if len(model_type.split(".")) > 0:
            model = model_type.split(".")[0]
        else:
            raise Exception("Invalid model_type.")
        if not hasattr(lightgbm, model):
            raise Exception("Invalid model_type.")

        model = getattr(lightgbm, model)

    elif model_type.startswith("catboost."):
        model_type = model_type.replace("catboost.", "")
        if len(model_type.split(".")) > 0:
            model = model_type.split(".")[0]
        else:
            raise Exception("Invalid model_type.")
        if not hasattr(catboost, model):
            raise Exception("Invalid model_type.")

        model = getattr(catboost, model)

    return model


def raise_data_exception():
    raise Exception("Cannot call this function without data.")


class MLModel(object):
    """
    Creates a machine learning model based on its implementation from scikit-learn, xgboost, lightgbm, catboost.
    For supervised learning:
    Refer `scikit-learn <https://scikit-learn.org/stable/supervised_learning.html#supervised-learning>`_,
    `xgboost <https://xgboost.readthedocs.io/en/stable/python/python_api.html>`_,
    `lightgbm <https://lightgbm.readthedocs.io/en/latest/Python-API.html>`_ ,
    `catboost <https://catboost.ai/en/docs/concepts/python-quickstart>`_ .

    For unsupervised learning:
    1. Clustering Models
    2. Gaussian Mixture Models
    3. Novelty and outlier detection
    Refer https://scikit-learn.org/stable/unsupervised_learning.html

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required TabularDataObject. Returned data object from
                            :class:`~arcgis.learn.prepare_tabulardata` function.
    ---------------------   -------------------------------------------
    model_type              Required string path to the module.
                            For example for SVM:

                            `sklearn.svm.SVR <https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVR.html>`_ or `sklearn.svm.SVC <https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html>`_

                            For tree:

                            `sklearn.tree.DecisionTreeRegressor <https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html>`_ or `sklearn.tree.DecisionTreeClassifier <https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html>`_

                            For gradient boosting:

                            `lightgbm.LGBMRegressor <https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMRegressor.html>`_ or `lightgbm.LGBMClassifier <https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html>`_
    ---------------------   -------------------------------------------
    Args:fairness_args(dict of str: str)        As of now we support only binary classification and Regression in fairness evaluation.

                                                A dictionary to provide fairness args. Following are allowed keys and values
                                                Keyword Args:                   Value Args:
                                                <sensitive_feature> str:        Protected class column or feature name. Ony categorical variable is allowed.
                                                <mitigation_type> str:          `reweighing` or `threshold_optimizer` or `exponentiated_gradient` (For Classification)
                                                                                `grid_search` or `exponentiated_gradient` (For Regression)
                                                <mitigation_constraint> str:    'demographic_parity' or'equalized_odds' (For Classification)
                                                                                and `ZeroOneLoss` or `SquareLoss` (For Regression)


                            For example:
                                                For classification :
                                                fairness_args = {
                                                            'sensitive_feature': 'Gender',
                                                            'mitigation_type': "threshold_optimizer",
                                                            'mitigation_constraint':'demographic_parity'

                                                            }

                                                For Regression :

                                                fairness_args = {
                                                            'sensitive_feature': 'Gender',
                                                            'mitigation_type': "grid_search",
                                                            'mitigation_constraint':'ZeroOneLoss'

                                                            }
    ---------------------   -------------------------------------------
    ``**kwargs``            model_type specific arguments.
                            Refer Parameters section

                            `scikit-learn <https://scikit-learn.org/stable/supervised_learning.html#supervised-learning>`_,

                            `lightgbm.LGBMRegressor <https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMRegressor.html>`_

                            `lightgbm.LGBMClassifier <https://lightgbm.readthedocs.io/en/latest/pythonapi/lightgbm.LGBMClassifier.html>`_

                            `catboostregressor <https://catboost.ai/en/docs/concepts/python-reference_catboostregressor>`_

                            `catboostclassifier <https://catboost.ai/en/docs/concepts/python-reference_catboostclassifier>`_

                            `xgboost <https://xgboost.readthedocs.io/en/stable/python/python_api.html#module-xgboost.sklearn>`_

    =====================   ===========================================

    :return: :class:`~arcgis.learn.MLModel` Object
    """

    # TODO Add fairness
    def __init__(self, data, model_type, fairness_args=None, **kwargs):
        if not HAS_ML_DEPS:
            raise Exception(missing_deps_trace)

        if data._cell_sizes:
            for res in data._cell_sizes:
                zone = f"zone{res}_id"
                if zone in data._field_mapping["categorical_variables"]:
                    data._field_mapping["categorical_variables"].remove(zone)
                if zone in data._dataframe:
                    data._dataframe = data._dataframe.drop(zone, axis=1)
                # data._categorical_variables.remove(zone)
            data._cell_sizes = None

        self._fairness = False
        self._model_type = model_type
        self._data = data

        if isinstance(self._data._dependent_variable, list):
            self._data._dependent_variable = self._data._dependent_variable[0]

        (
            self._training_data,
            self._training_labels,
            self._validation_data,
            self._validation_labels,
        ) = self._data._ml_data

        if kwargs.get("pretrained_model"):
            self._model = kwargs.get("pretrained_model")

            self._fairness = kwargs.get("fairness")
            self.mitigation_method = kwargs.get("mitigation_method")
            self.protected_class = kwargs.get("protected_class")

            if self._fairness and self._data._is_classification:
                self.fairness_label_encoder = LabelEncoder()
                self._training_labels = self.fairness_label_encoder.fit_transform(
                    self._training_labels
                )
                self._validation_labels = self.fairness_label_encoder.transform(
                    self._validation_labels
                )

        else:
            model = _get_model_type(model_type)

            if model == sklearn.cluster._kmeans.KMeans:
                if not kwargs.get("n_clusters"):
                    kwargs["n_clusters"] = self._get_number_of_clusters(**kwargs)
            elif model == sklearn.mixture._gaussian_mixture.GaussianMixture:
                if not kwargs.get("n_components"):
                    kwargs["n_components"] = self._get_number_of_components(**kwargs)

            self._model = model(**kwargs)

        self._training_df = pd.DataFrame(
            self._training_data,
            columns=self._data._continuous_variables
            + self._data._categorical_variables,
        )

        self._validation_df = pd.DataFrame(
            self._validation_data,
            columns=self._data._continuous_variables
            + self._data._categorical_variables,
        )

        if fairness_args is not None:
            self.initialize_fair_model(fairness_args)

    def initialize_fair_model(self, fairness_args):
        if not isinstance(fairness_args, dict):
            raise ValueError(_FAIRNESS_ARGS_NOT_DICT)

        if "sensitive_feature" not in fairness_args:
            raise ValueError(_FAIRNESS_ARGS_KEY_NOT_FOUND)

        self.protected_class = fairness_args["sensitive_feature"]

        if "mitigation_type" not in fairness_args:
            raise ValueError(_FAIRNESS_ARGS_KEY_NOT_FOUND)

        self.mitigation_method = fairness_args["mitigation_type"]

        self._fairness = True

        if self._data._is_classification:
            if "mitigation_constraint" not in fairness_args:
                self.constraint = "demographic_parity"
            else:
                self.constraint = fairness_args["mitigation_constraint"]

            if self.constraint not in ["demographic_parity", "equalized_odds"]:
                raise ValueError(_FAIRNESS_ARGS_KEY_NOT_FOUND)

            self.fairness_label_encoder = LabelEncoder()
            self._training_labels = self.fairness_label_encoder.fit_transform(
                self._training_labels
            )
            self._validation_labels = self.fairness_label_encoder.transform(
                self._validation_labels
            )

            self.instance_weights = None

            if self.mitigation_method == "reweighing":
                self._training_label_df = pd.DataFrame(
                    self._training_labels,
                    columns=[self._data._dependent_variable],
                )
                self._all_training_df = pd.concat(
                    [self._training_df, self._training_label_df], axis=1
                )

                self.instance_weights_train = _reweigh.assign_weights(
                    self._all_training_df,
                    self._data._dependent_variable,
                    self.protected_class,
                )

            elif self.mitigation_method == "exponentiated_gradient":
                if self.constraint == "demographic_parity":
                    mitigation_constraint = DemographicParity()
                elif self.constraint == "equalized_odds":
                    mitigation_constraint = EqualizedOdds()

                self._model = ExponentiatedGradient(
                    estimator=self._model,
                    constraints=mitigation_constraint,
                    sample_weight_name="sample_weight",
                )

            elif self.mitigation_method == "threshold_optimizer":
                if self.constraint not in [
                    "demographic_parity",
                    "selection_rate_parity",
                    "false_positive_rate_parity",
                    "true_negative_rate_parity",
                    "equalized_odds",
                ]:
                    raise ValueError(_FAIRNESS_ARGS_KEY_NOT_FOUND)

                self._model = ThresholdOptimizer(
                    estimator=self._model,
                    constraints=self.constraint,
                    objective="accuracy_score",
                    predict_method="predict_proba",
                )
            else:
                raise ValueError(_FAIRNESS_ARGS_KEY_NOT_FOUND)
        else:
            print("Initializing for Regression ")
            if "mitigation_constraint" not in fairness_args:
                self.constraint = "ZeroOneLoss"
            else:
                self.constraint = fairness_args["mitigation_constraint"]

            if self.constraint == "ZeroOneLoss":
                mitigation_constraint = ZeroOneLoss()
            elif self.constraint == "SquareLoss":
                mitigation_constraint = SquareLoss(min_val=0.001, max_val=0.1)
            else:
                raise ValueError(_FAIRNESS_ARGS_KEY_NOT_FOUND)

            if self.mitigation_method == "grid_search":
                self._model = GridSearch(
                    self._model,
                    constraints=BoundedGroupLoss(
                        mitigation_constraint, upper_bound=0.1
                    ),
                )
            elif self.mitigation_method == "exponentiated_gradient":
                self._model = ExponentiatedGradient(
                    estimator=self._model,
                    constraints=BoundedGroupLoss(
                        mitigation_constraint, upper_bound=0.1
                    ),
                    sample_weight_name="sample_weight",
                )
            else:
                raise ValueError(_FAIRNESS_ARGS_KEY_NOT_FOUND)

    def fit(self):
        if (
            not self._data._is_unsupervised
            and (self._training_data is None or self._training_labels is None)
        ) or (self._data._is_unsupervised and self._training_data is None):
            raise_data_exception()

        if self._data._is_unsupervised:
            self._model.fit(self._training_data)
        else:
            try:
                if self._fairness:
                    if self.mitigation_method == "reweighing":
                        print(f"Fitting with {self.mitigation_method}")
                        self._model.fit(
                            self._training_data,
                            self._training_labels,
                            sample_weight=self.instance_weights_train,
                        )
                    else:
                        try:
                            print(f"Fitting with {self.mitigation_method}")
                            self._model.fit(
                                self._training_df,
                                self._training_labels,
                                sensitive_features=self._training_df.loc[
                                    :, self.protected_class
                                ],
                            )
                        except ValueError as val:
                            val_error_message = val.args[0]
                            if (
                                "Degenerate labels for sensitive feature"
                                in val_error_message
                            ):
                                raise Exception(_DEGENERATE_LABEL_FOR_SENSITIVE_FEATURE)
                            else:
                                raise Exception(val)

                else:
                    self._model.fit(self._training_data, self._training_labels)

            except:
                raise Exception("Model is incompatible with the training data")

    def fairness_score(
        self,
        sensitive_feature,
        fairness_metrics=None,
        visualize=False,
    ):
        """
        Shows sample fairness score and plots for the model.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        sensitive_feature        Column name of the protected class.
        fairness_metrics         Allowed list of fairness metrics
                                 1. for classification
                                    [
                                     "equalized_odds_difference",
                                     "demographic_parity_difference",
                                     "equalized_odds_ratio",
                                     "demographic_parity_ratio"
                                    ]
                                 2. for Regression
                                    [
                                    "MAE",
                                    "MSE",
                                    "RMSE",
                                    "MAPE"
                                    ]
                                 Metric should be one of the values mentioned in
                                 the list.

        visualize                A boolean value to visualize plot of metrics
        =====================   ===========================================
        :return: dataframe
        """

        self.group_validation = self._validation_df.loc[:, [sensitive_feature]]
        if not self._fairness and self._data._is_classification:
            labelEncoder = LabelEncoder()
            train_labels = labelEncoder.fit_transform(self._training_labels)
            y_true = labelEncoder.transform(self._validation_labels)
            y_pred = self._predict(self._validation_df)

            y_pred = labelEncoder.transform(y_pred)
        else:
            y_true = self._validation_labels
            y_pred = self._predict(self._validation_df, self.group_validation)

        return _fairlearn.calculate_metrics(
            self._data._is_classification,
            self._data,
            y_true,
            y_pred,
            self.group_validation,
            sensitive_feature,
            fairness_metrics,
            visualize,
        )

    def show_results(self, rows=5):
        """
        Shows sample results for the model.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional number of rows. By default, 5 rows
                                are displayed.
        =====================   ===========================================
        :return: dataframe
        """
        if (
            not self._data._is_unsupervised
            and (self._validation_data is None or self._validation_labels is None)
        ) or (self._data._is_unsupervised and self._validation_data is None):
            raise_data_exception()

        min_size = len(self._validation_data)

        if rows < min_size:
            min_size = rows

        # sample_batch = random.sample(self._data._validation_indexes, min_size)
        sample_batch = random.sample(range(len(self._validation_data)), min_size)

        if self._fairness and self.mitigation_method == "threshold_optimizer":
            validation_df_batch = self._validation_df.iloc[sample_batch, :]
            sample_indexes = [self._data._validation_indexes[i] for i in sample_batch]
            group_df = validation_df_batch.loc[:, self.protected_class]
            output_labels = self._predict(validation_df_batch, group_df)
        else:
            validation_data_batch = self._validation_data.take(sample_batch, axis=0)
            sample_indexes = [self._data._validation_indexes[i] for i in sample_batch]
            output_labels = self._predict(validation_data_batch)

        pd.options.mode.chained_assignment = None

        # using loc instead of iloc to get data when dataframe doesnt have continuous indexes
        if self._data._is_classification:
            df = self._data._dataframe.loc[
                sample_indexes
            ]  # .loc[sample_batch]#.reset_index(drop=True).loc[sample_batch].reset_index(drop=True)
        else:
            df = self._data._dataframe.iloc[sample_indexes]

        if self._data._dependent_variable:
            df[self._data._dependent_variable + "_results"] = output_labels
        else:
            df["prediction_results"] = output_labels

        return df.sort_index()

    def score(self):
        """
        :return: output from scikit-learn's model.score(), R2 score in case of regression and Accuracy in case of classification.
        For KMeans returns Opposite of the value of X on the K-means objective.
        """
        if (
            not self._data._is_unsupervised
            and (self._validation_data is None or self._validation_labels is None)
        ) or (self._data._is_unsupervised and self._validation_data is None):
            raise_data_exception()

        if self._data._is_unsupervised:
            if hasattr(self._model, "score"):
                return self._model.score(self._training_data)

            raise Exception("Score function not applicable for unsupervised data")

        if self._fairness:
            if self.mitigation_method == "threshold_optimizer":
                group_df = self._validation_df.loc[:, self.protected_class]
                output_labels = self._predict(self._validation_df, group_df)
                return _fairlearn.score(
                    self._data._is_classification,
                    self._validation_labels,
                    output_labels,
                )
            else:
                return _fairlearn.score(
                    self._data._is_classification,
                    self._validation_labels,
                    self._predict(self._validation_df),
                )
        else:
            return self._model.score(self._validation_data, self._validation_labels)

    def decision_function(self):
        """
        :return: output from scikit-learn's model.decision_function()
        """
        if self._training_data is None:
            raise_data_exception()

        if not hasattr(self._model, "decision_function"):
            raise Exception("Function not implemented for this model.")

        return self._model.decision_function(self._training_data)

    def mahalanobis(self):
        """
        :return: output from scikit-learn's model.mahalanobis()
        """
        if self._training_data is None:
            raise_data_exception()

        if not hasattr(self._model, "mahalanobis"):
            raise Exception("Function not implemented for this model.")

        return self._model.mahalanobis(self._training_data)

    def kneighbors(self, X=None, n_neighbors=None, return_distance=True):
        """
        :return: output from scikit-learn's model.kneighbors()
        """
        if not hasattr(self._model, "kneighbors"):
            raise Exception("Function not implemented for this model.")

        kwargs = {}
        if X:
            kwargs["X"] = X
        elif self._training_data is None:
            raise Exception("No data found")
        else:
            kwargs["X"] = self._training_data

        if n_neighbors:
            kwargs["n_neighbors"] = n_neighbors

        kwargs["return_distance"] = return_distance

        return self._model.kneighbors(**kwargs)

    def predict_proba(self):
        """
        :return: output from scikit-learn's model.predict_proba()
        """

        if not hasattr(self._model, "predict_proba"):
            raise Exception("Function not implemented for this model.")

        if self._training_data is None:
            raise Exception("No data found.")

        return self._model.predict_proba(self._training_data)

    @property
    def feature_importances_(self):
        """
        :return: the global feature importance summary plot from SHAP. Most of the sklearn models are supported by this method.
        """
        # if not hasattr(self._model, 'feature_importances_'):
        # raise Exception("Property not implemented for this model.")
        processed_dataframe = None
        explain_index = None
        random_index = None
        explain_prediction(
            self,
            processed_dataframe,
            index=explain_index,
            random_index=random_index,
            predictor=None,
            global_pred=True,
        )
        return

        # return self._model.feature_importances_

    def save(self, name_or_path, publish=False, gis=None, **kwargs):
        """
        Saves the model, creates an Esri Model Definition. Uses pickle to save the model.
        Using protocol level 2. Protocol level is backward compatible.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Folder path to save the model.
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an item.
        ---------------------   -------------------------------------------
        gis                     Optional :class:`~arcgis.gis.GIS`  Object. Used for publishing the item.
                                If not specified then active gis user is taken.
        ---------------------   -------------------------------------------
        kwargs                  Optional Parameters:
                                Boolean `overwrite` if True, it will overwrite
                                the item on ArcGIS Online/Enterprise, default False.
        =====================   ===========================================
        :return: dataframe
        """

        if "\\" in name_or_path or "/" in name_or_path:
            path = name_or_path
        else:
            path = os.path.join(self._data.path, "models", name_or_path)
            if not os.path.exists(os.path.dirname(path)):
                os.mkdir(os.path.dirname(path))

        if not os.path.exists(os.path.dirname(path)):
            raise Exception("Path doesn't exist")

        if not os.path.exists(path):
            os.mkdir(path)

        base_file_name = os.path.basename(path)

        model_file = os.path.join(path, base_file_name + ".pkl")

        with open(model_file, "wb") as f:
            f.write(pickle.dumps(self._model, protocol=_PROTOCOL_LEVEL))

        MLModel._save_encoders(self._data._encoder_mapping, path, base_file_name)

        if self._data._procs:
            MLModel._save_transforms(self._data._procs, path, base_file_name)

        self._write_emd(path, base_file_name)
        zip_files = kwargs.pop("zip_files", True)

        if zip_files:
            from ._arcgis_model import _create_zip

            _create_zip(Path(path).name, str(path))

        if publish:
            file_name = os.path.basename(path) + ".dlpk"
            dlpk_path = Path(os.path.join(path, file_name))
            self._publish_dlpk(
                dlpk_path,
                gis=gis,
                overwrite=kwargs.get("overwrite", False),
            )

        return Path(path)

    def _publish_dlpk(self, dlpk_path, gis=None, overwrite=False):
        model_characteristics_folder = "ModelCharacteristics"
        gis_user = arcgis.env.active_gis if gis is None else gis
        if not gis_user:
            warnings.warn("No active gis user found!")
            return

        if not os.path.exists(dlpk_path):
            warnings.warn("DLPK file not found!")
            return

        emd_path = os.path.join(dlpk_path.parent, dlpk_path.stem + ".emd")

        if not os.path.exists(emd_path):
            warnings.warn("EMD File not found!")
            return

        emd_data = json.load(open(emd_path, "r"))
        formatted_description = f"""
                <p><b> {emd_data.get('ModelName').replace('>', '').replace('<', '')} </b></p>
                <p><b>Backbone:</b> {emd_data.get('ModelParameters', {}).get('backbone')}</p>
                <p><b>Learning Rate:</b> {emd_data.get('LearningRate')}</p>
        """

        if emd_data.get("accuracy"):
            formatted_description = (
                formatted_description
                + f"""
                <p><b>Analysis of the model</b></p>
                <p><b>Accuracy:</b> {emd_data.get('accuracy')}</p>
            """
            )

        if emd_data.get("average_precision_score"):
            formatted_description = (
                formatted_description
                + f"""
                <p><b>Analysis of the model</b></p>
                <p><b>Average Precision Score:</b> {emd_data.get('average_precision_score')}</p>
            """
            )

        item = gis_user.content.add(
            {
                "type": "Deep Learning Package",
                "description": formatted_description,
                "title": dlpk_path.stem,
                "overwrite": True if overwrite else False,
            },
            data=str(dlpk_path.absolute()),
        )

        print(f"Published DLPK Item Id: {item.itemid}")

    def _write_emd(self, path, base_file_name):
        emd_file = os.path.join(path, base_file_name + ".emd")
        emd_params = {}
        if self._model_type.startswith("sklearn."):
            emd_params["version"] = str(sklearn.__version__)
        elif self._model_type.startswith("xgboost."):
            emd_params["version"] = str(xgboost.__version__)
        elif self._model_type.startswith("lightgbm."):
            emd_params["version"] = str(lightgbm.__version__)
        elif self._model_type.startswith("catboost."):
            emd_params["version"] = str(catboost.__version__)
        else:
            emd_params["version"] = "Not Available"
        if not self._data._is_unsupervised:
            if self._data._is_empty:
                emd_params["score"] = self._data._emd["score"]
            else:
                emd_params["score"] = self.score()
        emd_params["_is_classification"] = (
            "classification" if self._data._is_classification else "regression"
        )
        emd_params["ModelName"] = type(self._model).__name__
        emd_params["ModelFile"] = base_file_name + ".pkl"

        if self._fairness:
            _mparams = self._model.get_params()
            _mparams["estimator"] = self._model_type
            _mparams["constraints"] = self.constraint
            emd_params["ModelParameters"] = _mparams
            emd_params["mitigation_method"] = self.mitigation_method
            emd_params["fairness"] = True
            emd_params["protected_class"] = self.protected_class

        else:
            emd_params["ModelParameters"] = self._model.get_params()
            emd_params["fairness"] = False

        emd_params["categorical_variables"] = self._data._categorical_variables

        if self._data._dependent_variable:
            emd_params["dependent_variable"] = self._data._dependent_variable

        emd_params["continuous_variables"] = self._data._continuous_variables
        emd_params["cell_sizes"] = self._data._cell_sizes

        with open(emd_file, "w") as f:
            f.write(json.dumps(emd_params, indent=4))

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.MLModel` Object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Esri Model Definition
                                file.
        ---------------------   -------------------------------------------
        data                    Required TabularDataObject or None. Returned data
                                object from :class:`~arcgis.learn.prepare_tabulardata` function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.MLModel` Object
        """
        if not HAS_ML_DEPS:
            raise Exception(missing_deps_trace)

        emd_path = str(emd_path)

        if emd_path.endswith(".dlpk"):
            with ZipFile(emd_path, "r") as zip_obj:
                temp_dir = tempfile.TemporaryDirectory().name
                zip_obj.extractall(temp_dir)
                MLModel.from_model(temp_dir, data)

        if not emd_path.endswith(".emd"):
            emd_path = os.path.join(
                emd_path, (str(os.path.basename(emd_path)) + ".emd")
            )

        if not os.path.exists(emd_path):
            raise Exception("Invalid data path.")

        with open(emd_path, "r") as f:
            emd = json.loads(f.read())

        categorical_variables = emd["categorical_variables"]
        dependent_variable = emd.get("dependent_variable", None)
        continuous_variables = emd["continuous_variables"]
        model_parameters = emd["ModelParameters"]
        model_parameters["fairness"] = emd["fairness"]
        fairness = model_parameters["fairness"]
        if fairness:
            model_parameters["mitigation_method"] = emd["mitigation_method"]
            model_parameters["protected_class"] = emd["protected_class"]

        cell_sizes = emd.get("cell_sizes", None)

        if (
            emd["version"] == str(sklearn.__version__)
            or emd["version"] == str(xgboost.__version__)
            or emd["version"] == str(lightgbm.__version__)
            or emd["version"] == str(catboost.__version__)
        ):
            pass
        else:
            warnings.warn(
                f"Sklearn/xgboost/lightgbm/catboost version has changed. Model Trained using version {emd['version']}"
            )

        _is_classification = True
        if emd["_is_classification"] != "classification":
            _is_classification = False

        encoder_mapping = None
        if categorical_variables:
            encoder_path = os.path.join(
                os.path.dirname(emd_path),
                os.path.basename(emd_path).split(".")[0] + "_encoders.pkl",
            )
            if os.path.exists(encoder_path):
                with open(encoder_path, "rb") as f:
                    encoder_mapping = pickle.loads(f.read())

        column_transformer = None
        transforms_path = os.path.join(
            os.path.dirname(emd_path),
            os.path.basename(emd_path).split(".")[0] + "_transforms.pkl",
        )
        if os.path.exists(transforms_path):
            with open(transforms_path, "rb") as f:
                column_transformer = pickle.loads(f.read())

        if data is None:
            data = TabularDataObject._empty(
                categorical_variables,
                continuous_variables,
                dependent_variable,
                encoder_mapping,
                column_transformer,
            )
            data._is_classification = _is_classification
            data._cell_sizes = cell_sizes

        data._emd = emd

        model_file = os.path.join(os.path.dirname(emd_path), emd["ModelFile"])
        with open(model_file, "rb") as f:
            model = pickle.loads(f.read())

        return cls(data, emd["ModelName"], pretrained_model=model, **model_parameters)

    def _predict(self, data, group_data=None):
        if self._fairness and self.mitigation_method == "threshold_optimizer":
            return self._model.predict(data, sensitive_features=group_data)
        return self._model.predict(data)

    @staticmethod
    def _save_encoders(encoder_mapping, path, base_file_name):
        if not encoder_mapping:
            return

        encoder_file = os.path.join(path, base_file_name + "_encoders.pkl")
        with open(encoder_file, "wb") as f:
            f.write(pickle.dumps(encoder_mapping, protocol=_PROTOCOL_LEVEL))

    @staticmethod
    def _save_transforms(column_transformer, path, base_file_name):
        if not column_transformer:
            return

        transforms_file = os.path.join(path, base_file_name + "_transforms.pkl")
        with open(transforms_file, "wb") as f:
            f.write(pickle.dumps(column_transformer, protocol=_PROTOCOL_LEVEL))

    @property
    def _is_kmeans(self):
        return self._model.__class__ == sklearn.cluster._kmeans.KMeans

    def _get_number_of_clusters(self, **kwargs):
        print("Finding optimum number of clusters")

        from sklearn.cluster import KMeans
        from sklearn.metrics import silhouette_score

        range_n_clusters = [
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            16,
            17,
            18,
            19,
            20,
        ]

        threshold = 5
        count = 0

        max_score = -1
        max_cluster = 2
        scores = []
        for n_clusters in range_n_clusters:
            # Initialize the clusterer with n_clusters value and a random generator
            # seed of 10 for reproducibility.
            clusterer = KMeans(n_clusters=n_clusters, **kwargs)
            cluster_labels = clusterer.fit_predict(self._training_data)

            # The silhouette_score gives the average value for all the samples.
            # This gives a perspective into the density and separation of the formed
            # clusters
            silhouette_avg = silhouette_score(self._training_data, cluster_labels)
            scores.append(silhouette_avg)

            if silhouette_avg > max_score:
                max_score = silhouette_avg
                max_cluster = n_clusters
                count = 0
            else:
                count = count + 1

            if count == threshold:
                break

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)

        ax.plot(range_n_clusters[0 : len(scores)], scores)
        ax.set_xlabel("Cluster")
        ax.set_ylabel("Silhouette scores")
        # ax.set_xscale('log')
        # ax.xaxis.set_major_formatter(plt.FormatStrFormatter('%.0e'))
        ax.plot(max_cluster, max_score, markersize=10, marker="o", color="red")

        plt.show()

        print(f"Selecting n_clusters={max_cluster}")

        return max_cluster

    def load(self, name_or_path):
        """
        Loads a compatible saved model for inferencing or fine tuning from the disk.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Name or Path to
                                Esri Model Definition(EMD) file.
        =====================   ===========================================
        """
        if not ("\\" in str(name_or_path) or "/" in str(name_or_path)):
            name_or_path = self._data.path / "models" / name_or_path
        model = MLModel.from_model(name_or_path, self._data)
        self._model = model._model

    def _get_number_of_components(self, **kwargs):
        print("Finding optimum number of components")

        from sklearn.mixture import GaussianMixture

        range_n_components = [
            2,
            3,
            4,
            5,
            6,
            7,
            8,
            9,
            10,
            11,
            12,
            13,
            14,
            15,
            16,
            17,
            18,
            19,
            20,
        ]

        max_score = -1
        max_component = 2

        threshold = 5
        count = 0
        scores = []
        for component in range_n_components:
            gm_model = GaussianMixture(n_components=component, **kwargs)

            gm_model.fit(self._training_data)
            score = gm_model.bic(self._training_data)

            scores.append(score)

            if score > max_score:
                max_component = component
                max_score = score
                count = 0
            else:
                count = count + 1

            if count == threshold:
                break

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(1, 1)

        ax.plot(range_n_components[0 : len(scores)], scores)
        ax.set_xlabel("Cluster")
        ax.set_ylabel("BIC score")
        # ax.set_xscale('log')
        # ax.xaxis.set_major_formatter(plt.FormatStrFormatter('%.0e'))
        ax.plot(max_component, max_score, markersize=10, marker="o", color="red")

        plt.show()

        print(f"Selecting n_components={max_component}")

        return max_component

    def predict(
        self,
        input_features=None,
        explanatory_rasters=None,
        datefield=None,
        distance_features=None,
        output_layer_name=None,
        gis=None,
        prediction_type="features",
        output_raster_path=None,
        match_field_names=None,
        explain=False,
        explain_index=None,
    ):
        """

        Predict on data from feature layer, dataframe and or raster data.

        =================================   =========================================================================
        **Parameter**                        **Description**
        ---------------------------------   -------------------------------------------------------------------------
        input_features                      Optional :class:`~arcgis.features.FeatureLayer` or spatial dataframe. Required if prediction_type='features'.
                                            Contains features with location and
                                            some or all fields required to infer the dependent variable value.
        ---------------------------------   -------------------------------------------------------------------------
        explanatory_rasters                 Optional list. Required if prediction_type='raster'.
                                            Contains a list of raster objects containing
                                            some or all fields required to infer the dependent variable value.
        ---------------------------------   -------------------------------------------------------------------------
        datefield                           Optional string. Field name from feature layer
                                            that contains the date, time for the input features.
                                            Same as :meth:`~arcgis.learn.prepare_tabulardata` .
        ---------------------------------   -------------------------------------------------------------------------
        distance_features                   Optional List of :class:`~arcgis.features.FeatureLayer` objects.
                                            These layers are used for calculation of field "NEAR_DIST_1",
                                            "NEAR_DIST_2" etc in the output dataframe.
                                            These fields contain the nearest feature distance
                                            from the input_features.
                                            Same as :meth:`~arcgis.learn.prepare_tabulardata` .
        ---------------------------------   -------------------------------------------------------------------------
        output_layer_name                   Optional string. Used for publishing the output layer.
        ---------------------------------   -------------------------------------------------------------------------
        gis                                 Optional :class:`~arcgis.gis.GIS`  Object. Used for publishing the item.
                                            If not specified then active gis user is taken.
        ---------------------------------   -------------------------------------------------------------------------
        prediction_type                     Optional String.
                                            Set 'features' or 'dataframe' to make output feature layer predictions.
                                            With this feature_layer argument is required.

                                            Set 'raster', to make prediction raster.
                                            With this rasters must be specified.
        ---------------------------------   -------------------------------------------------------------------------
        output_raster_path                  Optional path.
                                            Required when prediction_type='raster', saves
                                            the output raster to this path.
        ---------------------------------   -------------------------------------------------------------------------
        match_field_names                   Optional dictionary.
                                            Specify mapping of field names from prediction set
                                            to training set.
                                            For example:

                                                | {
                                                |    "Field_Name_1": "Field_1",
                                                |    "Field_Name_2": "Field_2"
                                                | }
        ---------------------------------   -------------------------------------------------------------------------
        explain                             Optional Bool.
                                            Setting this parameter to true generates prediction explanation plot.
                                            Plot is generated using model interpretability library called SHAP.
                                            (https://github.com/slundberg/shap)
        ---------------------------------   -------------------------------------------------------------------------
        explain_index                       Optional Int.
                                            The index of the dataframe passed to the predict function for which model
                                            interpretability is desired. If the parameter is not passed and if the
                                            explain parameter is set to true, the SHAP plot will be generated for a
                                            random index of the dataframe.
        =================================   =========================================================================

        :return: :class:`~arcgis.features.FeatureLayer` if prediction_type='features', dataframe for prediction_type='dataframe' else creates an output raster.

        """
        rasters = explanatory_rasters if explanatory_rasters else []
        if explain:
            try:
                import shap
            except:
                warnings.warn(
                    "Prediction cannot be explained as SHAP is not installed. Please install SHAP to get explainability working."
                )
                explain = False
                explain_index = None
        if prediction_type in ["features", "dataframe"]:
            if input_features is None:
                raise Exception("Feature Layer required for predict_features=True")

            gis = gis if gis else arcgis.env.active_gis
            return self._predict_features(
                input_features,
                rasters,
                datefield,
                distance_features,
                output_layer_name,
                gis,
                match_field_names,
                prediction_type,
                explain,
                explain_index,
            )
        else:
            if not rasters:
                raise Exception("Rasters required for predict_features=False")

            if not output_raster_path:
                raise Exception(
                    "Please specify output_raster_folder_path to save the output."
                )

            return self._predict_rasters(
                output_raster_path,
                rasters,
                match_field_names,
                explain,
                explain_index,
                output_layer_name,
                gis,
            )

    def _predict_features(
        self,
        input_features,
        rasters=None,
        datefield=None,
        distance_feature_layers=None,
        output_name="Prediction Layer",
        gis=None,
        match_field_names=None,
        prediction_type="features",
        explain=False,
        explain_index=None,
    ):
        if isinstance(input_features, FeatureLayer):
            dataframe = input_features.query().sdf
        else:
            dataframe = input_features.copy()

        fields_needed = (
            self._data._categorical_variables + self._data._continuous_variables
        )
        distance_feature_layers = (
            distance_feature_layers if distance_feature_layers else []
        )
        continuous_variables = self._data._continuous_variables

        columns = dataframe.columns
        feature_layer_columns = []
        for column in columns:
            column_name = column
            categorical = False

            if column_name in fields_needed:
                if column_name not in continuous_variables:
                    categorical = True
            elif match_field_names and match_field_names.get(column_name):
                if match_field_names.get(column_name) not in continuous_variables:
                    categorical = True
            else:
                continue

            feature_layer_columns.append((column_name, categorical))

        raster_columns = []
        if rasters:
            for raster in rasters:
                column_name = raster.name
                categorical = False
                if column_name in fields_needed:
                    if column_name not in continuous_variables:
                        categorical = True
                elif match_field_names and match_field_names.get(column_name):
                    column_name = match_field_names.get(column_name)
                    if column_name not in continuous_variables:
                        categorical = True
                else:
                    continue

                raster_columns.append((raster, categorical))

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            (
                processed_dataframe,
                fields_mapping,
            ) = TabularDataObject._prepare_dataframe_from_features(
                input_features,
                self._data._dependent_variable,
                feature_layer_columns,
                raster_columns,
                datefield,
                distance_feature_layers,
            )

        if match_field_names:
            processed_dataframe.rename(columns=match_field_names, inplace=True)

        for field in fields_needed:
            if field not in processed_dataframe.columns:
                raise Exception(f"Field missing {field}")

        for column in processed_dataframe.columns:
            if column not in fields_needed:
                processed_dataframe = processed_dataframe.drop(column, axis=1)

        processed_numpy = self._data._process_data(
            processed_dataframe.reindex(sorted(processed_dataframe.columns), axis=1),
            fit=False,
        )

        if self._fairness:
            processed_df = processed_numpy[self._validation_df.columns]
            group_data = processed_df.loc[:, self.protected_class]
            predictions = self._predict(processed_df, group_data)
        else:
            predictions = self._predict(processed_numpy)

        dataframe["prediction_results"] = predictions

        if self._fairness and self._data._is_classification:
            pred_transformed = self.fairness_label_encoder.inverse_transform(
                dataframe["prediction_results"]
            )
            dataframe["prediction_results"] = pred_transformed

        if explain:
            if explain_index is None:
                random_index = True
            else:
                random_index = False
            explain_prediction(
                self,
                processed_dataframe,
                index=explain_index,
                random_index=random_index,
                predictor=None,
                global_pred=False,
            )
        if prediction_type == "dataframe":
            return dataframe

        if "SHAPE" in list(dataframe.columns):
            return dataframe.spatial.to_featurelayer(output_name, gis)
        else:
            import tempfile

            with tempfile.TemporaryDirectory() as tmpdir:
                table_file = os.path.join(tmpdir, output_name + ".xlsx")
                dataframe.to_excel(table_file, index=False, header=True)
                online_table = gis.content.add(
                    {"type": "Microsoft Excel", "overwrite": True}, table_file
                )
                return online_table.publish(overwrite=True)

    def _predict_rasters(
        self,
        output_folder_path,
        rasters,
        match_field_names=None,
        explain=False,
        explain_index=None,
        output_layer_name=None,
        gis=None,
    ):
        if not os.path.exists(os.path.dirname(output_folder_path)):
            raise Exception("Output directory doesn't exist")

        if os.path.exists(output_folder_path):
            raise Exception("Output Folder already exists")

        try:
            import arcpy
        except:
            raise Exception("This function requires arcpy.")

        try:
            import numpy as np
        except:
            raise Exception("This function requires numpy.")

        try:
            import pandas as pd
        except:
            raise Exception("This function requires pandas.")

        if not HAS_FAST_PROGRESS:
            raise Exception("This function requires fastprogress.")

        fields_needed = (
            self._data._categorical_variables + self._data._continuous_variables
        )
        raster_names = [
            r.name for r in rasters
        ]  # Removing duplicate rasters if same raster is passed twice.
        unique_indexes = [raster_names.index(x) for x in set(raster_names)]
        rasters = [rasters[i] for i in unique_indexes]

        try:
            arcpy.env.outputCoordinateSystem = rasters[0].extent["spatialReference"][
                "wkt"
            ]
        except:
            arcpy.env.outputCoordinateSystem = rasters[0].extent["spatialReference"][
                "wkid"
            ]

        xmin = rasters[0].extent["xmin"]
        xmax = rasters[0].extent["xmax"]
        ymin = rasters[0].extent["ymin"]
        ymax = rasters[0].extent["ymax"]
        min_cell_size_x = rasters[0].mean_cell_width
        min_cell_size_y = rasters[0].mean_cell_height

        default_sr = rasters[0].extent["spatialReference"]

        for raster in rasters:
            point_upper = arcgis.geometry.Point(
                {
                    "x": raster.extent["xmin"],
                    "y": raster.extent["ymax"],
                    "sr": raster.extent["spatialReference"],
                }
            )
            point_lower = arcgis.geometry.Point(
                {
                    "x": raster.extent["xmax"],
                    "y": raster.extent["ymin"],
                    "sr": raster.extent["spatialReference"],
                }
            )
            cell_size = arcgis.geometry.Point(
                {
                    "x": raster.mean_cell_width,
                    "y": raster.mean_cell_height,
                    "sr": raster.extent["spatialReference"],
                }
            )

            points = arcgis.geometry.project(
                [point_upper, point_lower, cell_size],
                raster.extent["spatialReference"],
                default_sr,
            )
            point_upper = points[0]
            point_lower = points[1]
            cell_size = points[2]

            if xmin > point_upper.x:
                xmin = point_upper.x
            if ymax < point_upper.y:
                ymax = point_upper.y
            if xmax < point_lower.x:
                xmax = point_lower.x
            if ymin > point_lower.y:
                ymin = point_lower.y

            if min_cell_size_x > cell_size.x:
                min_cell_size_x = cell_size.x

            if min_cell_size_y > cell_size.y:
                min_cell_size_y = cell_size.y

        max_raster_columns = int(abs(math.ceil((xmax - xmin) / min_cell_size_x)))
        max_raster_rows = int(abs(math.ceil((ymax - ymin) / min_cell_size_y)))

        point_upper = arcgis.geometry.Point({"x": xmin, "y": ymax, "sr": default_sr})
        cell_size = arcgis.geometry.Point(
            {"x": min_cell_size_x, "y": min_cell_size_y, "sr": default_sr}
        )

        raster_data = {}
        # fields_needed.append(raster.name)# Change
        for raster in rasters:
            field_name = raster.name
            point_upper_translated = arcgis.geometry.project(
                [point_upper], default_sr, raster.extent["spatialReference"]
            )[0]
            cell_size_translated = arcgis.geometry.project(
                [cell_size], default_sr, raster.extent["spatialReference"]
            )[0]
            # if field_name in fields_needed:
            if field_name == re.search("(?s:.*)_", fields_needed[-1]).group(0)[:-1]:
                raster_read = raster.read(
                    origin_coordinate=(
                        point_upper_translated.x,
                        point_upper_translated.y,
                    ),
                    ncols=max_raster_columns,
                    nrows=max_raster_rows,
                    cell_size=(cell_size_translated.x, cell_size_translated.y),
                )
                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f"_{index}"
                            if not raster_data.get(key):
                                if key in fields_needed:
                                    raster_data[key] = []
                            index = index + 1
                            if key in fields_needed:
                                raster_data[key].append(value)
            elif match_field_names:
                field_name = list(match_field_names.values())[-1]
                field_name = re.search("(?s:.*)_", field_name).group(0)[:-1]
                # field_name = match_field_names.get(raster.name)
                raster_read = raster.read(
                    origin_coordinate=(
                        point_upper_translated.x,
                        point_upper_translated.y,
                    ),
                    ncols=max_raster_columns,
                    nrows=max_raster_rows,
                    cell_size=(cell_size_translated.x, cell_size_translated.y),
                )
                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f"_{index}"
                            if not raster_data.get(key):
                                if key in fields_needed:
                                    raster_data[key] = []
                            index = index + 1
                            if key in fields_needed:
                                raster_data[key].append(value)
            else:
                continue

        for field in fields_needed:
            if (
                field not in list(raster_data.keys())
                and match_field_names
                and match_field_names.get(field, None) is None
            ):
                raise Exception(f"Field missing {field}")

        processed_data = []

        length_values = len(raster_data[list(raster_data.keys())[0]])
        for i in range(length_values):
            processed_row = []
            for raster_name in sorted(raster_data.keys()):
                processed_row.append(raster_data[raster_name][i])
            processed_data.append(processed_row)

        processed_df = pd.DataFrame(
            data=np.array(processed_data), columns=sorted(raster_data)
        )

        processed_numpy = self._data._process_data(processed_df, fit=False)

        if explain:
            if explain_index is None:
                random_index = True
            else:
                random_index = False
            explain_prediction(
                self,
                processed_df,
                index=explain_index,
                random_index=random_index,
                predictor=None,
                global_pred=False,
            )

        predictions = self._predict(processed_numpy)

        predictions = np.array(
            predictions.reshape([max_raster_rows, max_raster_columns]), dtype="float64"
        )

        processed_raster = arcpy.NumPyArrayToRaster(
            predictions,
            arcpy.Point(xmin, ymin),
            x_cell_size=min_cell_size_x,
            y_cell_size=min_cell_size_y,
        )
        processed_raster.save(output_folder_path)
        if output_layer_name:
            from arcgis.raster.analytics import copy_raster

            copy_raster_op = copy_raster(
                input_raster=output_folder_path,
                output_name=output_layer_name.replace(" ", ""),
                raster_type_name="Raster Dataset",
                gis=gis,
                tiles_only=True,
            )
            print("The Published Item ID is : ", copy_raster_op.id)
            return
        return True



# Copyright 2020 The MMSegmentation Authors.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


_base_ = "./deeplabv3plus.py"
model = dict(
    pretrained="mmcls://mobilenet_v2",
    backbone=dict(
        _delete_=True,
        type="MobileNetV2",
        widen_factor=1.0,
        strides=(1, 2, 2, 1, 1, 1, 1),
        dilations=(1, 1, 1, 2, 2, 4, 4),
        out_indices=(1, 2, 4, 6),
    ),
    decode_head=dict(in_channels=320, c1_in_channels=24),
    auxiliary_head=dict(in_channels=96),
)

checkpoint = "https://download.openmmlab.com/mmsegmentation/v0.5/mobilenet_v2/deeplabv3plus_m-v2-d8_512x1024_80k_cityscapes/deeplabv3plus_m-v2-d8_512x1024_80k_cityscapes_20200825_124836-d256dd4b.pth"



from ._codetemplate import image_translation_prf
import json
import traceback
from .._data import _raise_fastai_import_error
from ._arcgis_model import ArcGISModel, _EmptyData

try:
    from ._wnet_cgan_utils import WNetcGANLoss, WNetcGANTrainer, optim, compute_metrics
    from ._wnet_cgan_utils import WNet_cGAN as wnet_cgan_model
    from .._utils.wnet_cgan import ImageTupleListMS2, _tensor_scaler_tfm, ImageTuple
    from .._utils.common import (
        get_multispectral_data_params_from_emd,
        _get_emd_path,
        ArcGISMSImage,
    )
    import torchvision
    from .._utils.wnet_cgan import show_results
    from pathlib import Path
    from fastai.vision import *
    from fastai.vision import DatasetType, Learner, partial, open_image
    import torch

    HAS_FASTAI = True
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False


class WNet_cGAN(ArcGISModel):

    """
    Creates a model object which generates images of type C from type A and type B.

    =====================   ===========================================
    **Parameter**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data` function.
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ===========================================

    :return:  :class:`~arcgis.learn.WNet_cGAN` Object
    """

    def __init__(self, data, pretrained_path=None, *args, **kwargs):
        super().__init__(data)
        self._check_dataset_support(data)
        if self._data.chip_size % 256 == 0:
            wnet_cgan_gan = wnet_cgan_model(self._data.n_channel, self._data.n_channel)
            self.learn = Learner(
                data,
                wnet_cgan_gan,
                loss_func=WNetcGANLoss(wnet_cgan_gan),
                opt_func=partial(optim.Adam, betas=(0.5, 0.99)),
                callback_fns=[WNetcGANTrainer],
            )
            self.learn.model = self.learn.model.to(self._device)
            self._slice_lr = False
            if pretrained_path is not None:
                self.load(pretrained_path)
            self._code = image_translation_prf

            def __str__(self):
                return self.__repr__()

            def __repr__(self):
                return "<%s>" % (type(self).__name__)

        else:
            raise Exception("Image chip sizes should be in multiples of 256")

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.WNet_cGAN` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :return:  :class:`~arcgis.learn.WNet_cGAN` Object
        """

        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        model_params = emd["ModelParameters"]
        resize_to = emd.get("resize_to")
        chip_size = emd["ImageHeight"]
        if data is None:
            data = _EmptyData(
                path=emd_path.parent, loss_func=None, c=2, chip_size=chip_size
            )
            data = get_multispectral_data_params_from_emd(data, emd)
            data._is_multispectral = emd.get("IsMultispectral", False)
            normalization_stats = dict(
                emd.get("NormalizationStats")
            )  # Copy the normalization stats so that self._data.emd has no tensors other wise it will raise error while creating emd
            for _stat in normalization_stats:
                if normalization_stats[_stat] is not None:
                    normalization_stats[_stat] = torch.tensor(
                        normalization_stats[_stat]
                    )
                setattr(data, ("_" + _stat), normalization_stats[_stat])
            normalization_stats_b = dict(emd.get("NormalizationStats_b"))
            for _stat in normalization_stats_b:
                if normalization_stats_b[_stat] is not None:
                    normalization_stats_b[_stat] = torch.tensor(
                        normalization_stats_b[_stat]
                    )
                setattr(data, ("_" + _stat + "_b"), normalization_stats_b[_stat])
            normalization_stats_c = dict(emd.get("NormalizationStats_c"))
            for _stat in normalization_stats_c:
                if normalization_stats_c[_stat] is not None:
                    normalization_stats_c[_stat] = torch.tensor(
                        normalization_stats_c[_stat]
                    )
                setattr(data, ("_" + _stat + "_c"), normalization_stats_c[_stat])
            data.n_channel = emd["n_channel"]
            data.nband_c = emd["n_band_c"]
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data.chip_size = chip_size
        data.resize_to = chip_size
        return cls(data, **model_params, pretrained_path=str(model_file))

    @property
    def _model_metrics(self):
        return self.compute_metrics(show_progress=True)

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_wnet_cgan"
        _emd_template["InferenceFunction"] = "ArcGISImageTranslation.py"
        _emd_template["ModelType"] = "WNet_cGAN"
        _emd_template["n_channel"] = self._data.n_channel
        _emd_template["n_band_a"] = self._data.nband_a
        _emd_template["n_band_b"] = self._data.nband_b
        _emd_template["n_band_c"] = self._data.nband_c
        _emd_template["NormalizationStats_c"] = {
            "band_min_values": self._data._band_min_values_c,
            "band_max_values": self._data._band_max_values_c,
            "band_mean_values": self._data._band_mean_values_c,
            "band_std_values": self._data._band_std_values_c,
            "scaled_min_values": self._data._scaled_min_values_c,
            "scaled_max_values": self._data._scaled_max_values_c,
            "scaled_mean_values": self._data._scaled_mean_values_c,
            "scaled_std_values": self._data._scaled_std_values_c,
        }
        for _stat in _emd_template["NormalizationStats_c"]:
            if _emd_template["NormalizationStats_c"][_stat] is not None:
                _emd_template["NormalizationStats_c"][_stat] = _emd_template[
                    "NormalizationStats_c"
                ][_stat].tolist()
        return _emd_template

    def show_results(self, rows=2, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================

        """
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)
        show_results(self, rows, **kwargs)

    def predict(self, img_path1, img_path2):
        """
        Predicts and display the image.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        img_path1                Required path of an image 1.
        img_path2                Required path of an image 2.
        =====================   ===========================================

        """

        self.learn.model.arcgis_results = True
        img_path1, img_path2 = Path(img_path1), Path(img_path2)
        n_band = self._data.n_channel

        raw_img1, raw_img2 = (
            ArcGISMSImage.open(img_path1),
            ArcGISMSImage.open(img_path2),
        )
        if n_band > raw_img1.shape[0]:
            last_tile = np.expand_dims(raw_img1.data[raw_img1.shape[0] - 1, :, :], 0)
            res = abs(n_band - raw_img1.shape[0])
            for i in range(res):
                raw_img1 = Image(
                    torch.tensor(np.concatenate((raw_img1.data, last_tile), axis=0))
                )
        if n_band > raw_img2.shape[0]:
            last_tile = np.expand_dims(raw_img2.data[raw_img2.shape[0] - 1, :, :], 0)
            res = abs(n_band - raw_img2.shape[0])
            for i in range(res):
                raw_img2 = Image(
                    torch.tensor(np.concatenate((raw_img2.data, last_tile), axis=0))
                )
        raw_img1_scaled = _tensor_scaler_tfm(
            raw_img1.data,
            min_values=self._data._band_min_values,
            max_values=self._data._band_max_values,
            mode="minmax",
        )
        raw_img2_scaled = _tensor_scaler_tfm(
            raw_img2.data,
            min_values=self._data._band_min_values_b,
            max_values=self._data._band_max_values_b,
            mode="minmax",
        )
        raw_img1_scaled_tensor = raw_img1_scaled[None].to(self._device)
        raw_img2_scaled_tensor = raw_img2_scaled[None].to(self._device)
        self.learn.model.eval()
        with torch.no_grad():
            prediction = (
                self.learn.model(
                    raw_img1_scaled_tensor,
                    raw_img2_scaled_tensor,
                    raw_img2_scaled_tensor,
                )[0]
                .detach()[0]
                .cpu()
            )
        pred_img = prediction / 2 + 0.5
        if self._data.nband_c == 1:
            pred_np = np.array(pred_img[0, :, :])
            pred_img = ArcGISMSImage(torch.tensor([pred_np] * pred_img.shape[0]))
        else:
            pred_img = ArcGISMSImage(pred_img)
        pred_img = pred_img.show()
        self.learn.model.arcgis_results = False
        return pred_img

    def compute_metrics(self, accuracy=True, show_progress=True):
        """
        Computes Peak Signal-to-Noise Ratio (PSNR) and
        Structural Similarity Index Measure (SSIM) on validation set.

        """

        psnr, ssim, ncc = compute_metrics(self, self._data.valid_dl, show_progress)
        # return {"PSNR":'{0:1.4e}'.format(psnr), "SSIM":'{0:1.4e}'.format(ssim), "NCC":'{0:1.4e}'.format(ncc)}
        return {"PSNR": psnr, "SSIM": ssim, "NCC": ncc}

    @property
    def supported_datasets(self):
        """Supported dataset types for this model."""
        return WNet_cGAN._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ["WNet_cGAN"]



from dataclasses import dataclass, asdict, field
from typing import Dict, Union, Optional, Any, ClassVar


@dataclass(init=False, frozen=True)
class _Unit:
    _distance: ClassVar[Dict[str, int]] = {
        "Kilometers": 9036,
        "Meters": 9001,
        "Centimeters": 1033,
        "Millimeters": 1025,
        "Fathoms": 9014,
        "Miles": 9035,
        "NauticalMiles": 9030,
        "Yards": 9096,
        "Feet": 9003,
        "Inches": 109009,
    }

    @staticmethod
    def _validate_distance(key: str) -> None:
        allowed_z_units = _Unit._distance.keys()
        if key not in allowed_z_units:
            raise ValueError(
                f"`{key}` is an invalid distance unit. Allowed values are - {', '.join(allowed_z_units)}"
            )


@dataclass(frozen=True)
class SingleFieldGeometry:
    """
    Dataclass that holds the Single Field Geometry configuration.

    ===============     ====================================================================
    **Parameter**        **Description**
    ===============     ====================================================================
    geometry_field      String. Geometry field name.

                        Options:

                            esriGeometryPoint, esriGeometryPolyline, esriGeometryPolygon, esriGeometryMulti.
    ---------------     --------------------------------------------------------------------
    geometry_type       String. Geometry type.

                        Options:

                            esriGeometryPoint, esriGeometryPolyline, esriGeometryPolygon, esriGeometryMulti.
    ---------------     --------------------------------------------------------------------
    geometry_format     String. Geometry format.

                        Options:

                            coordinates, esrijson, geojson, or wkt.
    ---------------     --------------------------------------------------------------------
    wkid                int. WKID of the geometry.
    ===============     ====================================================================

    :return: `True` if the operation is a success

    .. code-block:: python

        # Usage Example

        geometry = SingleFieldFeometry(
            geometry_field="geometry_field"
            geometry_type="esriGeometryPoint",
            geometry_format="esrijson",
            wkid=4326
        )

    """

    geometry_field: str
    geometry_type: str
    geometry_format: str
    wkid: int

    def __post_init__(self):
        allowed_geometry_types = [
            "esriGeometryPoint",
            "esriGeometryPolyline",
            "esriGeometryPolygon",
            "esriGeometryMulti",
        ]
        if self.geometry_type not in allowed_geometry_types:
            raise ValueError(
                f"`{self.geometry_type}` is an invalid geometry_type. Allowed values are - {', '.join(allowed_geometry_types)}"
            )

        allowed_geometry_formats = ["coordinates", "esrijson", "geojson", "wkt"]
        if self.geometry_format not in allowed_geometry_formats:
            raise ValueError(
                f"`{self.geometry_format}` is an invalid gemetry_type. Allowed values are - {', '.join(allowed_geometry_formats)}"
            )


@dataclass(frozen=True)
class XYZGeometry:
    """
    Dataclass that holds the XYZ Geometry configuration.

    =====================   ====================================================================
    **Parameter**            **Description**
    ---------------------   --------------------------------------------------------------------
    x_field                 String. Longitude field name.
    ---------------------   --------------------------------------------------------------------
    y_field                 String. Latitude field name.
    ---------------------   --------------------------------------------------------------------
    wkid                    int. WKID of the geometry.
    =====================   ====================================================================

    =====================   ====================================================================
    **Optional Argument**   **Description**
    =====================   ====================================================================
    z_field                 String. Z field name.
    ---------------------   --------------------------------------------------------------------
    z_unit                  String. Z units. Options: Kilometers, Meters, Centimeters, Millimeters, Fathoms,
                            Miles, NauticalMiles, Yards, Feet, Inches.
    =====================   ====================================================================

    :return: `True` if the operation is a success

    .. code-block:: python

        # Usage Example

        geometry = XYZGeometry(
            x_field = "x",
            y_field = "y",
            wkid = 4326
        )

    """

    x_field: str
    y_field: str
    wkid: int
    z_field: str = None
    z_unit: str = None

    def __post_init__(self):
        if self.z_field is not None:
            _Unit._validate_distance(self.z_unit)


class _HasGeometry:
    # ---> Inheriting classes MUST declare the following Optional fields. <---
    # These are commented variables because of a limitation in dataclass where base class properties get ordered before
    # the derived class properties. Since these are optional properties, the init will order them before the non-optional
    # derived class properties leading to errors like - TypeError: non-default argument 'rss_url' follows default argument
    #
    # geometry: Optional[Union[XYZGeometry, SingleFieldGeometry]] = None

    def set_geometry_config(
        self, geometry: Union[XYZGeometry, SingleFieldGeometry]
    ) -> bool:
        """
        Configures the geometry for a feed.

        ==============     ====================================================================
        **Parameter**        **Description**
        ---------------     --------------------------------------------------------------------
        geometry            [:class:`~arcgis.realtime.velocity.feeds.XYZGeometry`, :class:`~arcgis.realtime.velocity.feeds.SingleFieldGeometry`].
                            Geometry object used to configure the feed.
        ===============     ====================================================================

        :return: `True` if the operation is a success

        .. code-block:: python

        # Usage Example

        feed.set_geometry_config(geometry=geometry)

        """
        if isinstance(geometry, XYZGeometry):
            self.data_format.x_field = geometry.x_field
            self.data_format.y_field = geometry.y_field
            if geometry.z_field is not None:
                self.data_format.has_z_field = True
                self.data_format.z_field = geometry.z_field
                if geometry.z_unit:
                    _Unit._validate_distance(geometry.z_unit)
                    z_unit_code = _Unit._distance.get(geometry.z_unit)
                    self.data_format.z_unit = z_unit_code
            self.data_format.build_geometry_from_fields = True

            self._fields["geometry"] = {
                "geometryType": "esriGeometryPoint",
                "spatialReference": {"wkid": geometry.wkid},
                "hasZ": self.data_format.has_z_field,
            }

            return True
        elif isinstance(geometry, SingleFieldGeometry):
            self.data_format.geometry_field = geometry.geometry_field
            self.data_format.geometry_field_format = geometry.geometry_format
            self.data_format.build_geometry_from_fields = False

            self._fields["geometry"] = {
                "geometryType": geometry.geometry_type,
                "spatialReference": {"wkid": geometry.wkid},
            }

            return True



from typing import Union, Dict, Any, Optional, ClassVar
from dataclasses import field, dataclass

from arcgis.realtime import Velocity
from arcgis.realtime.velocity.feeds._feed_template import _FeedTemplate
from arcgis.realtime.velocity.feeds.run_interval import RunInterval
from arcgis.realtime.velocity.feeds.time import _HasTime, TimeInstant, TimeInterval


@dataclass
class FeatureLayer(_FeedTemplate, _HasTime):
    """
    Poll a feature layer for features at a fixed schedule. This data class can be used to define the feed configuration
    and to create the feed.

    The data format is a feature layer. ArcGIS Velocity will automatically handle the location for you.

    =====================           ====================================================================
    **Parameter**                    **Description**
    ---------------------           --------------------------------------------------------------------
    label                           String. Unique label for this feed instance.
    ---------------------           --------------------------------------------------------------------
    description                     String. Feed description.
    ---------------------           --------------------------------------------------------------------
    query                           String. Feature layer query parameters. The default is: 1=1.
    ---------------------           --------------------------------------------------------------------
    fields                          String. Requested feature layer output fields.

                                    For example:

                                           "field1, field2"

                                    The default is: *.
    ---------------------           --------------------------------------------------------------------
    outSR                           int. Requested output spatial reference. The default is: 4326.

                                    .. note::
                                        To learn more about projected and geographic coordinate systems, refer to
                                        `Using spatial references <https://developers.arcgis.com/rest/services-reference/enterprise/using-spatial-references.htm>`_.
    =====================           ====================================================================

    =====================           ====================================================================
    **Optional Argument**           **Description**
    =====================           ====================================================================
    portal_item_id                  String. The Portal :class:`~arcgis.gis.Item` ID of the feature layer.

                                    .. note::
                                        Either the portal_item_id or url is required.
    ---------------------           --------------------------------------------------------------------
    extent                          Dict[str, Any]. A Geometry object that defines the spatial extent for
                                    the feature layer.

                                    .. code-block:: python

                                        # Sample Value
                                        {
                                            "spatialReference": {
                                                "latestWkid": 3857,
                                                "wkid": 102100
                                            },
                                            "xmin": -14784278.027601289,
                                            "ymin": 2604610.848073723,
                                            "xmax": -11451317.846255329,
                                            "ymax": 6852675.132049575
                                        }

    ---------------------           --------------------------------------------------------------------
    time_stamp_field                String.
                                    An optional date field for latest features.
                                    Optionally, specify a date field to be used to retrieve only the latest
                                    features from the feature layer.

                                    If a timestamp field is not specified, ArcGIS Velocity will load all
                                    features that meet the criteria of the WHERE clause when it polls the feature layer.

                                    If a timestamp field is specified, the first time ArcGIS Velocity polls the feature
                                    layer it will load all features with a timestamp field datetime within the past
                                    minute and less than the first feed poll time that also meets the criteria of the
                                    WHERE clause. With each subsequent poll, only features with a timestamp field value
                                    between the last polling time and the current polling time that also meet the
                                    criteria of the WHERE clause will be loaded.
    ---------------------           --------------------------------------------------------------------
    track_id_field                  String. Name of the field from the incoming data that should be set as
                                    track ID.
    ---------------------           --------------------------------------------------------------------
    time                            [:class:`~arcgis.realtime.velocity.feeds.TimeInstant`, :class:`~arcgis.realtime.velocity.feeds.TimeInterval`]. An instance of time configuration that
                                    will be used to create time information from the incoming data.
    ---------------------           --------------------------------------------------------------------
    run_interval                    :class:`~arcgis.realtime.velocity.feeds.RunInterval`. An instance of the scheduler configuration. The default is:
                                    RunInterval(cron_expression="0 * * ? * * *", timezone="America/Los_Angeles")
    =====================           ====================================================================

    :return: A data class with feature layer feed configuration.

    .. code-block:: python

        # Usage Example

        from arcgis.realtime.velocity.feeds import FeatureLayer
        from arcgis.realtime.velocity.http_authentication_type import (
            NoAuth,
            BasicAuth,
            CertificateAuth,
        )

        from arcgis.realtime.velocity.input.format import DelimitedFormat
        from arcgis.realtime.velocity.feeds.geometry import XYZGeometry, SingleFieldGeometry
        from arcgis.realtime.velocity.feeds.time import TimeInterval, TimeInstant
        from arcgis.realtime.velocity.feeds.run_interval import RunInterval

        extent = {
            "spatialReference": {
                "latestWkid": 3857,
                "wkid": 102100
            },
            "xmin": "xmin",
            "ymin": "ymin",
            "xmax": "xmax",
            "ymax": "ymax"
        }

        # Feature Layer Properties

        feature_layer_config = FeatureLayer(
            label="feed_name",
            description="feed_description",
            query="1=1",
            fields="*",
            outSR=4326,
            url="feed_sample_server_link",
            extent=extent,
            time_stamp_field="date_field"
        )

        feature_layer_config

        # Set recurrence
        feature_layer_config.run_interval = RunInterval(
            cron_expression="0 * * ? * * *", timezone="America/Los_Angeles"
        )

        # use velocity object to get the FeedsManager instance
        feeds = velocity.feeds

        # use the FeedsManager object to create a feed from this feed configuration
        feature_layer_feed = feeds.create(feature_layer_config)
        feature_layer_feed.start()
        feeds.items

    """

    # fields that the user sets during init
    # Feature Layer specific properties
    query: str = field(default="1=1")
    fields: str = field(default="*")
    outSR: int = field(default=4326)
    url: Optional[str] = None
    portal_item_id: Optional[str] = None
    extent: Optional[Dict[str, Any]] = None
    time_stamp_field: Optional[str] = None

    # FeedTemplate properties
    track_id_field: Optional[str] = None
    # HasTime properties
    time: Optional[Union[TimeInstant, TimeInterval]] = None
    # scheduler
    run_interval: RunInterval = field(
        default=RunInterval(
            cron_expression="0 * * ? * * *", timezone="America/Los_Angeles"
        )
    )
    # Feature Layer is a standard format and format properties do not need to be set in the feed configuration
    data_format: Any = field(default=None, init=False)
    # FeedTemplate properties
    _name: ClassVar[str] = "feature-layer"

    def __post_init__(self):
        if Velocity is None:
            return
        self._util = Velocity._util

        # validation of fields
        if self._util.is_valid(self.label) == False:
            raise ValueError(
                "Label should only contain alpha numeric, _ and space only"
            )
        if self.url is None and self.portal_item_id is None:
            raise ValueError("Either url or portal_item_id is required")

        # generate dictionary of this feed object's properties that will be used to query test-connection and
        # sample-messages Rest endpoint
        feed_properties = self._generate_feed_properties()

        test_connection = self._util.test_connection(
            input_type="feed", payload=feed_properties
        )
        # Test connection to make sure feed can fetch schema
        if test_connection is True:
            # test connection succeeded. Now try getting sample messages
            sample_payload = {
                "properties": {
                    "maxSamplesToCollect": 5,
                    "timeoutInMillis": 5000,
                }
            }
            self._dict_deep_merge(sample_payload, feed_properties)
            # Sample messages to fetch schema/fields
            sample_messages = self._util.sample_messages(
                input_type="feed", payload=sample_payload
            )

            if sample_messages["featureSchema"] is not None:
                # sample messages succeeded. Get Feature Schema from it
                self._set_fields(sample_messages["featureSchema"])

            print(
                "Feature Schema retrieved from the Feed:",
                sample_messages["featureSchema"],
            )

            # initiate actions for each of the following properties if it was set at init
            if self.track_id_field is not None:
                self.set_track_id(self.track_id_field)
            if self.time is not None:
                self.set_time_config(self.time)

    def _build(self) -> dict:
        feed_configuration = {
            "id": "",
            "label": self.label,
            "description": self.description,
            "feed": {**self._generate_schema_transformation()},
            **self.run_interval._build(),
            "properties": {"executable": True},
        }

        feed_properties = self._generate_feed_properties()
        self._dict_deep_merge(feed_configuration["feed"], feed_properties)
        print(feed_configuration)
        return feed_configuration

    def _generate_feed_properties(self) -> dict:
        if self.url:
            url_or_portal_item_id = {f"{self._name}.url": self.url}
        else:
            url_or_portal_item_id = {f"{self._name}.portalItemId": self.portal_item_id}

        if self.extent:
            extent_properties = {f"{self._name}.extent": self.extent}
        else:
            extent_properties = {}

        if self.time_stamp_field:
            time_stamp_property = {
                f"{self._name}.timestampField": self.time_stamp_field
            }
        else:
            time_stamp_property = {}

        feed_properties = {
            "name": self._name,
            "properties": {
                f"{self._name}.query": self.query,
                f"{self._name}.fields": self.fields,
                f"{self._name}.outSR": self.outSR,
                **url_or_portal_item_id,
                **extent_properties,
                **time_stamp_property,
            },
        }

        return feed_properties



# model settings
custom_imports = dict(imports=["arcgis.learn.models._prithvi_archs"])

# required bands Blue, Green, Red, Narrow NIR, SWIR 1, SWIR 2
bands = [0, 1, 2, 3, 4, 5]

CLASSES = ("Unburnt land", "Burn scar")

img_norm_burn_model = dict(
    means=[
        0.033349706741586264,
        0.05701185520536176,
        0.05889748132001316,
        0.2323245113436119,
        0.1972854853760658,
        0.11944914225186566,
    ],
    stds=[
        0.02269135568823774,
        0.026807560223070237,
        0.04004109844362779,
        0.07791732423672691,
        0.08708738838140137,
        0.07241979477437814,
    ],
)

norm_cfg = dict(type="BN", requires_grad=True)
model = dict(
    type="TemporalEncoderDecoder",
    pretrained=None,
    frozen_backbone=True,
    backbone=dict(
        type="TemporalViTEncoder",
        pretrained=None,
        img_size=224,
        patch_size=16,
        num_frames=1,
        tubelet_size=1,
        in_chans=len(bands),
        embed_dim=768,
        depth=12,
        num_heads=12,
        mlp_ratio=4.0,
        norm_pix_loss=False,
    ),
    neck=dict(
        type="ConvTransformerTokensToEmbeddingNeck",
        embed_dim=768 * 1,
        output_embed_dim=768 * 1,
        drop_cls_token=True,
        Hp=14,
        Wp=14,
    ),
    decode_head=dict(
        in_channels=768 * 1,
        type="FCNHead",
        in_index=-1,
        channels=256,
        num_convs=1,
        concat_input=False,
        dropout_ratio=0.1,
        norm_cfg=dict(type="BN", requires_grad=True),
        align_corners=False,
        loss_decode=dict(
            type="DiceLoss", use_sigmoid=False, loss_weight=1, ignore_index=-1
        ),
    ),
    auxiliary_head=dict(
        in_channels=768 * 1,
        type="FCNHead",
        in_index=-1,
        channels=256,
        num_convs=2,
        concat_input=False,
        dropout_ratio=0.1,
        norm_cfg=dict(type="BN", requires_grad=True),
        align_corners=False,
        loss_decode=dict(
            type="DiceLoss", use_sigmoid=False, loss_weight=1, ignore_index=-1
        ),
    ),
    # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict(mode="whole"),
)

checkpoint = "https://huggingface.co/ibm-nasa-geospatial/Prithvi-100M-burn-scar/resolve/main/burn_scars_Prithvi_100M.pth"



import json
from arcgis._impl.common._mixins import PropertyMap
from arcgis._impl.common._isd import InsensitiveDict


class _Metadata(object):
    """
    Internal `Metadata` that stores information about the source data
    """

    _source = None
    _renderer = None

    # ----------------------------------------------------------------------
    def __init__(self):
        """initializer"""
        pass

    # ----------------------------------------------------------------------
    @property
    def source_type(self):
        if self.source is None:
            return None
        return self.source.__class__.__name__

    # ----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__}>"

    # ----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    # ----------------------------------------------------------------------
    @property
    def source(self):
        """gets/sets the source data pointer"""
        return self._source

    # ----------------------------------------------------------------------
    @source.setter
    def source(self, value):
        """gets/sets the source data pointer"""
        self._source = value

    # ----------------------------------------------------------------------
    @property
    def renderer(self):
        """gets/sets the renderer"""
        return self._renderer

    # ----------------------------------------------------------------------
    @renderer.setter
    def renderer(self, value):
        """gets/sets the renderer"""
        if isinstance(value, dict):
            value = InsensitiveDict.from_dict(value)
        elif isinstance(value, PropertyMap):
            value = InsensitiveDict.from_dict(dict(value))
        elif isinstance(value, InsensitiveDict):
            pass
        else:
            raise ValueError("Must be a a dictionary or InsensitiveDict")
        if value != self._renderer:
            self._renderer = value

    # ----------------------------------------------------------------------
    def __setstate__(self, d):
        """unpickle support"""
        self.__dict__.update(d)

    # ----------------------------------------------------------------------
    def __getstate__(self):
        """pickle support"""
        return self.__dict__



from ._lyrs import Object3DLayer, IntegratedMeshLayer, Point3DLayer
from ._lyrs import PointCloudLayer, BuildingLayer, VoxelLayer
from ._lyrs import SceneLayer, SceneLayerManager, EnterpriseSceneLayerManager



from __future__ import annotations
import json
from arcgis.gis import Layer, _GISResource, Item
from arcgis.geoprocessing import import_toolbox
from arcgis.geometry import Geometry
from arcgis.auth.tools import LazyLoader

_services = LazyLoader("arcgis.gis.server.admin._services")


class SceneLayerManager(_GISResource):
    """
    The ``SceneLayerManager`` class allows administration (if access permits) of ArcGIS Online hosted scene layers.
    A :class:`~arcgis.mapping.SceneLayerManager` offers access to map and layer content.
    """

    def __init__(self, url, gis=None, scene_lyr=None):
        if url.split("/")[-1].isdigit():
            url = url.replace(f"/{url.split('/')[-1]}", "")
        super(SceneLayerManager, self).__init__(url, gis)
        self._sl = scene_lyr
        # Scene Layers published from Scene Layer Package are read only.
        if "layers" in self.properties:
            self._source_type = (
                "Feature Service"
                if "updateEnabled" in self.properties.layers[0]
                else "Scene Layer Package"
            )
        else:
            # No layers are present so we will not have cache
            self._source_type = "Scene Layer Package"

    # ----------------------------------------------------------------------
    def refresh(self):
        """
        The ``refresh`` operation refreshes a service, which clears the web
        server cache for the service.
        """
        if self._source_type == "Scene Layer Package":
            url = self._url + "SceneServer/refresh"
            params = {"f": "json"}

            res = self._con.post(url, params)

            super(SceneLayerManager, self)._refresh()

            return res
        return None

    # ----------------------------------------------------------------------
    def swap(self, target_service_name: str):
        """
        The swap operation replaces the current service cache with an existing one.

        .. note::
            The ``swap`` operation is for ArcGIS Online only.

        ====================        ====================================================
        **Parameter**                **Description**
        --------------------        ----------------------------------------------------
        target_service_name         Required string. Name of service you want to swap with.
        ====================        ====================================================

        :returns: dictionary indicating success or error

        """
        if self._source_type == "Scene Layer Package":
            url = self._url + "/swap"
            params = {"f": "json", "targetServiceName": target_service_name}
            return self._con.post(url, params)
        return None

    # ----------------------------------------------------------------------
    def jobs(self):
        """
        The tile service job summary (jobs) resource represents a
        summary of all jobs associated with a vector tile service.
        Each job contains a jobid that corresponds to the specific
        jobid run and redirects you to the Job Statistics page.

        """
        if self._source_type == "Scene Layer Package":
            url = self._url + "/jobs"
            params = {"f": "json"}
            return self._con.get(url, params)
        return None

    # ----------------------------------------------------------------------
    def cancel_job(self, job_id: str):
        """
        The ``cancel_job`` operation supports cancelling a job while update
        tiles is running from a hosted feature service. The result of this
        operation is a response indicating success or failure with error
        code and description.

        ===============     ====================================================
        **Parameter**        **Description**
        ---------------     ----------------------------------------------------
        job_id              Required String. The job id to cancel.
        ===============     ====================================================

        """
        if self._source_type == "Scene Layer Package":
            url = self._url + "/jobs/%s/cancel" % job_id
            params = {"f": "json"}
            return self._con.post(url, params)
        return None

    # ----------------------------------------------------------------------
    def job_statistics(self, job_id: str):
        """
        Returns the job statistics for the given jobId

        """
        if self._source_type == "Scene Layer Package":
            url = self._url + "/jobs/%s" % job_id
            params = {"f": "json"}
            return self._con.post(url, params)
        return None

    # -----------------------------------------------------------------------
    def rerun_job(self, job_id: str, code: str):
        """
        The ``rerun_job`` operation supports re-running a canceled job from a
        hosted map service. The result of this operation is a response
        indicating success or failure with error code and description.

        ===============     ====================================================
        **Parameter**        **Description**
        ---------------     ----------------------------------------------------
        code                Required string, parameter used to re-run a given
                            jobs with a specific error
                            code: ``ALL | ERROR | CANCELED``
        ---------------     ----------------------------------------------------
        job_id              Required string, job to reprocess
        ===============     ====================================================

        :return:
           A boolean or dictionary
        """
        if self._source_type == "Scene Layer Package":
            url = self._url + "/jobs/%s/rerun" % job_id
            params = {"f": "json", "rerun": code}
            return self._con.post(url, params)
        return None

    # ----------------------------------------------------------------------
    def import_package(self, item: str | Item):
        """
        The ``import`` method imports from an :class:`~arcgis.gis.Item` object.

        ===============     ====================================================
        **Parameter**        **Description**
        ---------------     ----------------------------------------------------
        item                Required ItemId or :class:`~arcgis.gis.Item` object. The TPK file's item id.
                            This TPK file contains to-be-extracted bundle files
                            which are then merged into an existing cache service.
        ===============     ====================================================

        :return:
            A dictionary

        """
        if self._source_type == "Scene Layer Package":
            params = {
                "f": "json",
                "sourceItemId": None,
            }
            if isinstance(item, str):
                params["sourceItemId"] = item
            elif isinstance(item, Item):
                params["sourceItemId"] = item.itemid
            else:
                raise ValueError("The `item` must be a string or Item")
            url = self._url + "/import"
            res = self._con.post(url, params)
            return res
        return None

    # ----------------------------------------------------------------------
    def update(self):
        """
        The ``update`` method starts update generation for ArcGIS Online. It updates
        the underlying source dataset for the service, essentially refreshing the
        underlying package data.

        :return:
           Dictionary.
        """
        if self._gis._portal.is_arcgisonline:
            url = "%s/update" % self._url
            params = {"f": "json"}
            return self._con.post(url, params)
        return None

    # ----------------------------------------------------------------------
    def edit(self, item: str | Item):
        """
        The ``edit`` method edits from an :class:`~arcgis.gis.Item` object.

        ===============     ====================================================
        **Parameter**        **Description**
        ---------------     ----------------------------------------------------
        item                Required ItemId or :class:`~arcgis.gis.Item` object. The TPK file's item id.
                            This TPK file contains to-be-extracted bundle files
                            which are then merged into an existing cache service.
        ===============     ====================================================

        :return:
            A dictionary

        """
        if self._source_type == "Scene Layer Package":
            params = {
                "f": "json",
                "sourceItemId": None,
            }
            if isinstance(item, str):
                params["sourceItemId"] = item
            elif isinstance(item, Item):
                params["sourceItemId"] = item.itemid
            else:
                raise ValueError("The `item` must be a string or Item")
            url = self._url + "/edit"
            res = self._con.post(url, params)
            return res
        return None

    # ----------------------------------------------------------------------
    def rebuild_cache(self, layers: int | list[int]):
        """
        The rebuild_cache operation update the scene layer cache to reflect
        any changes made to the feature layer used to publish this scene layer.
        The results of the operation is a response indicating success, which
        redirects you to the Job Statistics page, or failure.

        =====================       ====================================================
        **Parameter**                **Description**
        ---------------------       ----------------------------------------------------
        layers                      Required int or list of int. Comma seperated values indicating
                                    the id of the layers to rebuild in the cache.

                                    Ex: [0,1,2]
        =====================       ====================================================
        """
        if self._source_type == "Feature Service":
            url = self._url + "/rebuildCache"
            params = {"f": "json", "layers": layers}
            return self._con.post(url, params)
        return None

    # ----------------------------------------------------------------------
    def update_cache(self, layers: int | list[int]):
        """
        Update Cache is a "light rebuild" where attributes and geometries of
        the layers selected are updated and can be used for change tracking on
        the feature layer to only update nodes with dirty tiles.
        The results of the operation is a response indicating success, which
        redirects you to the Job Statistics page, or failure.

        =====================       ====================================================
        **Parameter**                **Description**
        ---------------------       ----------------------------------------------------
        layers                      Required int or list of int. Comma seperated values indicating
                                    the id of the layers to update in the cache.

                                    Ex: [0,1,2]
        =====================       ====================================================
        """
        if self._source_type == "Feature Service":
            url = self._url + "/updateCache"
            params = {"f": "json", "layers": layers}
            return self._con.post(url, params)
        return None

    # ----------------------------------------------------------------------
    def update_attribute(self, layers: int | list[int]):
        """
        Update atrribute is a "light rebuild" where attributes of
        the layers selected are updated and can be used for change tracking.
        The results of the operation is a response indicating success, which
        redirects you to the Job Statistics page, or failure.

        =====================       ====================================================
        **Parameter**                **Description**
        ---------------------       ----------------------------------------------------
        layers                      Required int or list of int. Comma seperated values indicating
                                    the id of the layers to update in the cache.

                                    Ex: [0,1,2]
        =====================       ====================================================
        """
        if self._source_type == "Feature Service":
            url = self._url + "/updateAttribute"
            params = {"f": "json", "layers": layers}
            return self._con.post(url, params)
        return None


###########################################################################
class EnterpriseSceneLayerManager(_GISResource):
    """
    The ``EnterpriseSceneLayerManager`` class allows administration (if access permits) of ArcGIS Enterprise hosted scene layers.
    A :class:`~arcgis.mapping.SceneLayer` offers access to layer content.

    .. note:: Url must be admin url such as: ``https://services.myserver.com/arcgis/rest/admin/services/serviceName/SceneServer/``
    """

    _gptbx = None

    def __init__(self, url, gis=None, scene_lyr=None):
        if url.split("/")[-1].isdigit():
            url = url.replace(f"/{url.split('/')[-1]}", "")
        super(EnterpriseSceneLayerManager, self).__init__(url, gis)
        self._sl = scene_lyr

    # ----------------------------------------------------------------------
    def edit(self, service_dictionairy: dict):
        """
        To edit a service, you need to submit the complete JSON
        representation of the service, which includes the updates to the
        service properties. Editing a service causes the service to be
        restarted with updated properties.

        ===================     ====================================================================
        **Parameter**            **Description**
        -------------------     --------------------------------------------------------------------
        service_dictionairy     Required dict. The service JSON as a dictionary.
        ===================     ====================================================================


        :return: boolean
        """
        sl_service = _services.Service(self.url, self._gis)
        return sl_service.edit(service_dictionairy)

    # ----------------------------------------------------------------------
    def start(self):
        """starts the specific service"""
        sl_service = _services.Service(self.url, self._gis)
        return sl_service.start()

    # ----------------------------------------------------------------------
    def stop(self):
        """stops the specific service"""
        sl_service = _services.Service(self.url, self._gis)
        return sl_service.stop()

    # ----------------------------------------------------------------------
    def change_provider(self, provider: str):
        """
        Allows for the switching of the service provide and how it is hosted on the ArcGIS Server instance.

        Values:

        + 'ArcObjects' means the service is running under the ArcMap runtime i.e. published from ArcMap
        + 'ArcObjects11': means the service is running under the ArcGIS Pro runtime i.e. published from ArcGIS Pro
        + 'DMaps': means the service is running in the shared instance pool (and thus running under the ArcGIS Pro provider runtime)

        :return: Boolean

        """
        sl_service = _services.Service(self.url, self._gis)
        return sl_service.change_provider(provider)

    # ----------------------------------------------------------------------
    def delete(self):
        """deletes a service from arcgis server"""
        sl_service = _services.Service(self.url, self._gis)
        return sl_service.delete()

    # ----------------------------------------------------------------------
    @property
    def _tbx(self):
        """gets the toolbox"""
        if self._gptbx is None:
            self._gptbx = import_toolbox(
                url_or_item=self._gis.hosting_servers[0].url
                + "/System/SceneCachingControllers/GPServer",
                gis=self._gis,
            )
            self._gptbx._is_fa = True
        return self._gptbx

    # ----------------------------------------------------------------------
    def rebuild_cache(
        self,
        layer: list[int] | None = None,
        extent: dict | None = None,
        area_of_interest: dict | None = None,
    ):
        """
        The rebuild_cache operation update the scene layer cache to reflect
        any changes made to the feature layer used to publish this scene layer.
        The results of the operation is the url to the scene service once it is
        done rebuilding.

        ===============================     ====================================================================
        **Parameter**                        **Description**
        -------------------------------     --------------------------------------------------------------------
        layer                               Optional list of integers. The list of layers to cook.
        -------------------------------     --------------------------------------------------------------------
        extent                              Optional dict. The updated extent to be used. If nothing is specified,
                                            the default extent is used.
        -------------------------------     --------------------------------------------------------------------
        area_of_interest                    Optional dict representing a feature. Specify the updated area
                                            of interest.

                                            Syntax:
                                                {
                                                    "displayFieldName": "",
                                                    "geometryType": "esriGeometryPolygon",
                                                    "spatialReference": {
                                                    "wkid": 54051,
                                                    "latestWkid": 54051
                                                    },
                                                    "fields": [
                                                    {
                                                    "name": "OID",
                                                    "type": "esriFieldTypeOID",
                                                    "alias": "OID"
                                                    },
                                                    {
                                                    "name": "updateGeom_Length",
                                                    "type": "esriFieldTypeDouble",
                                                    "alias": "updateGeom_Length"
                                                    },
                                                    {
                                                    "name": "updateGeom_Area",
                                                    "type": "esriFieldTypeDouble",
                                                    "alias": "updateGeom_Area"
                                                    }
                                                    ],
                                                    "features": [],
                                                    "exceededTransferLimit": False
                                                }
        ===============================     ====================================================================

        :return: If successful, the url to the scene service

        """
        if layer is None:
            layer = {}
        elif layer is not None:
            layer = {layer}
        if extent is None:
            extent = "DEFAULT"
        if area_of_interest is None:
            return self._tbx.manage_scene_cache(
                service_url=self._sl.url,
                num_of_caching_service_instances=2,
                layer=layer,
                update_mode="RECREATE_ALL_NODES",
                update_extent=extent,
            )
        else:
            return self._tbx.manage_scene_cache(
                service_url=self._sl.url,
                num_of_caching_service_instances=2,
                layer=layer,
                update_mode="RECREATE_ALL_NODES",
                update_extent=extent,
                area_of_interest=area_of_interest,
            )

    # ----------------------------------------------------------------------
    def update_cache(
        self,
        layer: list[int] | None = None,
        extent: dict | None = None,
        area_of_interest: dict | None = None,
    ):
        """
        Update Cache is a "light rebuild" where attributes and geometries of
        the layers selected are updated and can be used for change tracking on
        the feature layer to only update nodes with dirty tiles,.
        The results of the operation is the url to the scene service once it is
        done updating.

        ===============================     ====================================================================
        **Parameter**                        **Description**
        -------------------------------     --------------------------------------------------------------------
        layer                               Optional list of integers. The list of layers to cook.
        -------------------------------     --------------------------------------------------------------------
        extent                              Optional dict. The updated extent to be used. If nothing is specified,
                                            the default extent is used.
        -------------------------------     --------------------------------------------------------------------
        area_of_interest                    Optional dict representing a feature. Specify the updated area
                                            of interest.

                                            Syntax:
                                                {
                                                    "displayFieldName": "",
                                                    "geometryType": "esriGeometryPolygon",
                                                    "spatialReference": {
                                                    "wkid": 54051,
                                                    "latestWkid": 54051
                                                    },
                                                    "fields": [
                                                    {
                                                    "name": "OID",
                                                    "type": "esriFieldTypeOID",
                                                    "alias": "OID"
                                                    },
                                                    {
                                                    "name": "updateGeom_Length",
                                                    "type": "esriFieldTypeDouble",
                                                    "alias": "updateGeom_Length"
                                                    },
                                                    {
                                                    "name": "updateGeom_Area",
                                                    "type": "esriFieldTypeDouble",
                                                    "alias": "updateGeom_Area"
                                                    }
                                                    ],
                                                    "features": [],
                                                    "exceededTransferLimit": False
                                                }
        ===============================     ====================================================================

        :return: If successful, the url to the scene service

        """
        if layer is None:
            layer = {}
        elif layer is not None:
            layer = {layer}
        if extent is None:
            extent = "DEFAULT"
        if area_of_interest is None:
            return self._tbx.manage_scene_cache(
                service_url=self._sl.url,
                num_of_caching_service_instances=2,
                layer=layer,
                update_mode="PARTIAL_UPDATE_NODES",
                update_extent=extent,
            )
        else:
            return self._tbx.manage_scene_cache(
                service_url=self._sl.url,
                num_of_caching_service_instances=2,
                layer=layer,
                update_mode="PARTIAL_UPDATE_NODES",
                update_extent=extent,
                area_of_interest=area_of_interest,
            )

    # ----------------------------------------------------------------------
    def update_attribute(
        self,
        layer: list[int] | None = None,
        extent: dict | None = None,
        area_of_interest: dict | None = None,
    ):
        """
        Update atrribute is a "light rebuild" where attributes of
        the layers selected are updated and can be used for change tracking.
        The results of the operation is the url to the scene service once it is
        done updating.

        ===============================     ====================================================================
        **Parameter**                        **Description**
        -------------------------------     --------------------------------------------------------------------
        layer                               Optional list of integers. The list of layers to cook.
        -------------------------------     --------------------------------------------------------------------
        extent                              Optional dict. The updated extent to be used. If nothing is specified,
                                            the default extent is used.
        -------------------------------     --------------------------------------------------------------------
        area_of_interest                    Optional dict representing a feature. Specify the updated area
                                            of interest.

                                            Syntax:
                                                {
                                                    "displayFieldName": "",
                                                    "geometryType": "esriGeometryPolygon",
                                                    "spatialReference": {
                                                    "wkid": 54051,
                                                    "latestWkid": 54051
                                                    },
                                                    "fields": [
                                                    {
                                                    "name": "OID",
                                                    "type": "esriFieldTypeOID",
                                                    "alias": "OID"
                                                    },
                                                    {
                                                    "name": "updateGeom_Length",
                                                    "type": "esriFieldTypeDouble",
                                                    "alias": "updateGeom_Length"
                                                    },
                                                    {
                                                    "name": "updateGeom_Area",
                                                    "type": "esriFieldTypeDouble",
                                                    "alias": "updateGeom_Area"
                                                    }
                                                    ],
                                                    "features": [],
                                                    "exceededTransferLimit": false
                                                }
        ===============================     ====================================================================

        :return: If successful, the url to the scene service
        """
        if layer is None:
            layer = {}
        elif layer is not None:
            layer = {layer}
        if extent is None:
            extent = "DEFAULT"
        if area_of_interest is None:
            return self._tbx.manage_scene_cache(
                service_url=self._sl.url,
                num_of_caching_service_instances=2,
                layer=layer,
                update_mode="PARTIAL_UPDATE_ATTRIBUTES",
                update_extent=extent,
            )
        else:
            return self._tbx.manage_scene_cache(
                service_url=self._sl.url,
                num_of_caching_service_instances=2,
                layer=layer,
                update_mode="PARTIAL_UPDATE_ATTRIBUTES",
                update_extent=extent,
                area_of_interest=area_of_interest,
            )


###########################################################################
class Object3DLayer(Layer):
    """
    The ``Object3DLayer`` represents a Web scene 3D Object layer.

    .. note::
        Web scene layers are cached web layers that are optimized for displaying a large amount of 2D and 3D features.
        See the :class:`~arcgis.mapping.SceneLayer` class for more information.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional :class:`~arcgis.gis.GIS`  object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.Point3DLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """

    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(Object3DLayer, self).__init__(url, gis)
        self._admin = None

    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict = {"type": "SceneLayer", "url": url}
        if self._token is not None:
            lyr_dict["serviceToken"] = self._token

        if self.filter is not None:
            lyr_dict["filter"] = self.filter
        if self._time_filter is not None:
            lyr_dict["time"] = self._time_filter
        return lyr_dict

    # ----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += "?token=" + self._token

        lyr_dict = {"type": "SceneLayer", "url": url}

        if self.filter is not None:
            lyr_dict["options"] = json.dumps({"definition_expression": self.filter})
        if self._time_filter is not None:
            lyr_dict["time"] = self._time_filter
        return lyr_dict

    # ----------------------------------------------------------------------
    @property
    def manager(self):
        """
        The ``manager`` property returns an instance of :class:`~arcgis.mapping.SceneLayerManager` class
        or :class:`~arcgis.mapping.EnterpriseSceneLayerManager` class
        which provides methods and properties for administering this service.
        """
        if self._admin is None:
            if self._gis._portal.is_arcgisonline:
                rd = {"/rest/services/": "/rest/admin/services/"}
                adminURL = self._str_replace(self._url, rd)
                if adminURL.split("/")[-1].isdigit():
                    adminURL = adminURL.replace(f'/{adminURL.split("/")[-1]}', "")
                self._admin = SceneLayerManager(adminURL, self._gis, self)
            else:
                rd = {"/rest/": "/admin/", "/SceneServer": ".SceneServer"}
                adminURL = self._str_replace(self._url, rd)
                if adminURL.split("/")[-1].isdigit():
                    adminURL = adminURL.replace(f'/{adminURL.split("/")[-1]}', "")
                self._admin = EnterpriseSceneLayerManager(adminURL, self._gis, self)
        return self._admin

    # ----------------------------------------------------------------------
    def _str_replace(self, mystring, rd):
        """Replaces a value based on a key/value pair where the
        key is the text to replace and the value is the new value.

        The find/replace is case insensitive.

        """
        import re

        patternDict = {}
        for key, value in rd.items():
            pattern = re.compile(re.escape(key), re.IGNORECASE)
            patternDict[value] = pattern
        for key in patternDict:
            regex_obj = patternDict[key]
            mystring = regex_obj.sub(key, mystring)
        return mystring


###########################################################################
class IntegratedMeshLayer(Layer):
    """
    The ``IntegratedMeshLayer`` class represents a Web scene Integrated Mesh layer.

    .. note::
        Web scene layers are cached web layers that are optimized for displaying a large amount of 2D and 3D features.
        See the :class:`~arcgis.mapping.SceneLayer` class for more information.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional :class:`~arcgis.gis.GIS`  object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.Point3DLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """

    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(IntegratedMeshLayer, self).__init__(url, gis)
        self._admin = None

    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict = {"type": "IntegratedMeshLayer", "url": url}
        if self._token is not None:
            lyr_dict["serviceToken"] = self._token

        if self.filter is not None:
            lyr_dict["filter"] = self.filter
        if self._time_filter is not None:
            lyr_dict["time"] = self._time_filter
        return lyr_dict

    # ----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += "?token=" + self._token

        lyr_dict = {"type": "IntegratedMeshLayer", "url": url}

        if self.filter is not None:
            lyr_dict["options"] = json.dumps({"definition_expression": self.filter})
        if self._time_filter is not None:
            lyr_dict["time"] = self._time_filter
        return lyr_dict

    # ----------------------------------------------------------------------
    @property
    def manager(self):
        """
        The ``manager`` property returns an instance of :class:`~arcgis.mapping.SceneLayerManager` class
        or :class:`~arcgis.mapping.EnterpriseSceneLayerManager` class
        which provides methods and properties for administering this service.
        """
        if self._admin is None:
            if self._gis._portal.is_arcgisonline:
                rd = {"/rest/services/": "/rest/admin/services/"}
                adminURL = self._str_replace(self._url, rd)
                if adminURL.split("/")[-1].isdigit():
                    adminURL = adminURL.replace(f'/{adminURL.split("/")[-1]}', "")
                self._admin = SceneLayerManager(adminURL, self._gis, self)
            else:
                rd = {"/rest/": "/admin/", "/SceneServer": ".SceneServer"}
                adminURL = self._str_replace(self._url, rd)
                if adminURL.split("/")[-1].isdigit():
                    adminURL = adminURL.replace(f'/{adminURL.split("/")[-1]}', "")
                self._admin = EnterpriseSceneLayerManager(adminURL, self._gis, self)
        return self._admin

    # ----------------------------------------------------------------------
    def _str_replace(self, mystring, rd):
        """Replaces a value based on a key/value pair where the
        key is the text to replace and the value is the new value.

        The find/replace is case insensitive.

        """
        import re

        patternDict = {}
        for key, value in rd.items():
            pattern = re.compile(re.escape(key), re.IGNORECASE)
            patternDict[value] = pattern
        for key in patternDict:
            regex_obj = patternDict[key]
            mystring = regex_obj.sub(key, mystring)
        return mystring


###########################################################################


class VoxelLayer(Layer):
    """
    The ``VoxelLayer`` class represents a Web Scene Voxel layer.

    .. note::
        Web scene layers are cached web layers that are optimized for displaying
        a large amount of 2D and 3D features. See the
        :class:`~arcgis.mapping.SceneLayer` class for more information.

    ==================     =============================================================
    **Parameter**           **Description**
    ------------------     -------------------------------------------------------------
    url                    Required string, specify the url ending in ``/SceneServer/``
    ------------------     -------------------------------------------------------------
    gis                    Optional :class:`~arcgis.gis.GIS` object. If not specified,
                           the active GIS connection is used.
    ==================     =============================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.VoxelLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """

    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(VoxelLayer, self).__init__(url, gis)
        self._admin = None

    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict = {"type": "VoxelLayer", "url": url}
        if self._token is not None:
            lyr_dict["serviceToken"] = self._token

        if self.filter is not None:
            lyr_dict["filter"] = self.filter
        if self._time_filter is not None:
            lyr_dict["time"] = self._time_filter
        return lyr_dict

    # ----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += "?token=" + self._token

        lyr_dict = {"type": "VoxelLayer", "url": url}

        if self.filter is not None:
            lyr_dict["options"] = json.dumps({"definition_expression": self.filter})
        if self._time_filter is not None:
            lyr_dict["time"] = self._time_filter
        return lyr_dict

    # ----------------------------------------------------------------------
    @property
    def manager(self):
        """
        The ``manager`` property returns an instance of
        :class:`~arcgis.mapping.SceneLayerManager` class
        or :class:`~arcgis.mapping.EnterpriseSceneLayerManager` class
        which provides methods and properties for administering this service.
        """
        if self._admin is None:
            if self._gis._portal.is_arcgisonline:
                rd = {"/rest/services/": "/rest/admin/services/"}
                adminURL = self._str_replace(self._url, rd)
                if adminURL.split("/")[-1].isdigit():
                    adminURL = adminURL.replace(f'/{adminURL.split("/")[-1]}', "")
                self._admin = SceneLayerManager(adminURL, self._gis, self)
            else:
                rd = {"/rest/": "/admin/", "/SceneServer": ".SceneServer"}
                adminURL = self._str_replace(self._url, rd)
                if adminURL.split("/")[-1].isdigit():
                    adminURL = adminURL.replace(f'/{adminURL.split("/")[-1]}', "")
                self._admin = EnterpriseSceneLayerManager(adminURL, self._gis, self)
        return self._admin

    # ----------------------------------------------------------------------
    def _str_replace(self, mystring, rd):
        """Replaces a value based on a key/value pair where the
        key is the text to replace and the value is the new value.

        The find/replace is case insensitive.

        """
        import re

        patternDict = {}
        for key, value in rd.items():
            pattern = re.compile(re.escape(key), re.IGNORECASE)
            patternDict[value] = pattern
        for key in patternDict:
            regex_obj = patternDict[key]
            mystring = regex_obj.sub(key, mystring)
        return mystring


###########################################################################
class Point3DLayer(Layer):
    """
    The ``Point3DLayer`` class represents a Web scene 3D Point layer.

    .. note::
        Web scene layers are cached web layers that are optimized for displaying a large amount of 2D and 3D features.
        See the :class:`~arcgis.mapping.SceneLayer` class for more information.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional :class:`~arcgis.gis.GIS`  object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.Point3DLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """

    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(Point3DLayer, self).__init__(url, gis)
        self._admin = None

    # ----------------------------------------------------------------------
    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict = {"type": "SceneLayer", "url": url}
        if self._token is not None:
            lyr_dict["serviceToken"] = self._token

        if self.filter is not None:
            lyr_dict["filter"] = self.filter
        if self._time_filter is not None:
            lyr_dict["time"] = self._time_filter
        return lyr_dict

    # ----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += "?token=" + self._token

        lyr_dict = {"type": "SceneLayer", "url": url}

        if self.filter is not None:
            lyr_dict["options"] = json.dumps({"definition_expression": self.filter})
        if self._time_filter is not None:
            lyr_dict["time"] = self._time_filter
        return lyr_dict

    # ----------------------------------------------------------------------
    @property
    def manager(self):
        """
        The ``manager`` property returns an instance of :class:`~arcgis.mapping.SceneLayerManager` class
        or :class:`~arcgis.mapping.EnterpriseSceneLayerManager` class
        which provides methods and properties for administering this service.
        """
        if self._admin is None:
            if self._gis._portal.is_arcgisonline:
                rd = {"/rest/services/": "/rest/admin/services/"}
                adminURL = self._str_replace(self._url, rd)
                if adminURL.split("/")[-1].isdigit():
                    adminURL = adminURL.replace(f'/{adminURL.split("/")[-1]}', "")
                self._admin = SceneLayerManager(adminURL, self._gis, self)
            else:
                rd = {"/rest/": "/admin/", "/SceneServer": ".SceneServer"}
                adminURL = self._str_replace(self._url, rd)
                if adminURL.split("/")[-1].isdigit():
                    adminURL = adminURL.replace(f'/{adminURL.split("/")[-1]}', "")
                self._admin = EnterpriseSceneLayerManager(adminURL, self._gis, self)
        return self._admin

    # ----------------------------------------------------------------------
    def _str_replace(self, mystring, rd):
        """Replaces a value based on a key/value pair where the
        key is the text to replace and the value is the new value.

        The find/replace is case insensitive.

        """
        import re

        patternDict = {}
        for key, value in rd.items():
            pattern = re.compile(re.escape(key), re.IGNORECASE)
            patternDict[value] = pattern
        for key in patternDict:
            regex_obj = patternDict[key]
            mystring = regex_obj.sub(key, mystring)
        return mystring


###########################################################################
class PointCloudLayer(Layer):
    """
    The ``PointCloudLayer`` class represents a Web scene Point Cloud layer.

    .. note::
        Point Cloud layers are cached web layers that are optimized for displaying a large amount of 2D and 3D features.
        See the :class:`~arcgis.mapping.SceneLayer` class for more information.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional :class:`~arcgis.gis.GIS`  object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.PointCloudLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """

    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(PointCloudLayer, self).__init__(url, gis)
        self._admin = None

    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict = {"type": "PointCloudLayer", "url": url}
        if self._token is not None:
            lyr_dict["serviceToken"] = self._token

        if self.filter is not None:
            lyr_dict["filter"] = self.filter
        if self._time_filter is not None:
            lyr_dict["time"] = self._time_filter
        return lyr_dict

    # ----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += "?token=" + self._token

        lyr_dict = {"type": "PointCloudLayer", "url": url}

        if self.filter is not None:
            lyr_dict["options"] = json.dumps({"definition_expression": self.filter})
        if self._time_filter is not None:
            lyr_dict["time"] = self._time_filter
        return lyr_dict

    # ----------------------------------------------------------------------
    @property
    def manager(self):
        """
        The ``manager`` property returns an instance of :class:`~arcgis.mapping.SceneLayerManager` class
        or :class:`~arcgis.mapping.EnterpriseSceneLayerManager` class
        which provides methods and properties for administering this service.
        """
        if self._admin is None:
            if self._gis._portal.is_arcgisonline:
                rd = {"/rest/services/": "/rest/admin/services/"}
                adminURL = self._str_replace(self._url, rd)
                if adminURL.split("/")[-1].isdigit():
                    adminURL = adminURL.replace(f'/{adminURL.split("/")[-1]}', "")
                self._admin = SceneLayerManager(adminURL, self._gis, self)
            else:
                rd = {"/rest/": "/admin/", "/SceneServer": ".SceneServer"}
                adminURL = self._str_replace(self._url, rd)
                if adminURL.split("/")[-1].isdigit():
                    adminURL = adminURL.replace(f'/{adminURL.split("/")[-1]}', "")
                self._admin = EnterpriseSceneLayerManager(adminURL, self._gis, self)
        return self._admin

    # ----------------------------------------------------------------------
    def _str_replace(self, mystring, rd):
        """Replaces a value based on a key/value pair where the
        key is the text to replace and the value is the new value.

        The find/replace is case insensitive.

        """
        import re

        patternDict = {}
        for key, value in rd.items():
            pattern = re.compile(re.escape(key), re.IGNORECASE)
            patternDict[value] = pattern
        for key in patternDict:
            regex_obj = patternDict[key]
            mystring = regex_obj.sub(key, mystring)
        return mystring


###########################################################################
class BuildingLayer(Layer):
    """
    The ``BuildingLayer`` class represents a Web building layer.

    .. note::
        Web scene layers are cached web layers that are optimized for displaying a large amount of 2D and 3D features.
        See the :class:`~arcgis.mapping.SceneLayer` class for more information.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional :class:`~arcgis.gis.GIS`  object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.BuildingLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """

    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(BuildingLayer, self).__init__(url, gis)
        self._admin = None

    @property
    def _lyr_dict(self):
        url = self.url

        lyr_dict = {"type": "BuildingSceneLayer", "url": url}
        if self._token is not None:
            lyr_dict["serviceToken"] = self._token

        if self.filter is not None:
            lyr_dict["filter"] = self.filter
        if self._time_filter is not None:
            lyr_dict["time"] = self._time_filter
        return lyr_dict

    # ----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        url = self.url
        if self._token is not None:  # causing geoanalytics Invalid URL error
            url += "?token=" + self._token

        lyr_dict = {"type": "BuildingSceneLayer", "url": url}

        if self.filter is not None:
            lyr_dict["options"] = json.dumps({"definition_expression": self.filter})
        if self._time_filter is not None:
            lyr_dict["time"] = self._time_filter
        return lyr_dict

    # ----------------------------------------------------------------------
    @property
    def manager(self):
        """
        The ``manager`` property returns an instance of :class:`~arcgis.mapping.SceneLayerManager` class
        or :class:`~arcgis.mapping.EnterpriseSceneLayerManager` class
        which provides methods and properties for administering this service.
        """
        if self._admin is None:
            if self._gis._portal.is_arcgisonline:
                rd = {"/rest/services/": "/rest/admin/services/"}
                adminURL = self._str_replace(self._url, rd)
                if adminURL.split("/")[-1].isdigit():
                    adminURL = adminURL.replace(f'/{adminURL.split("/")[-1]}', "")
                self._admin = SceneLayerManager(adminURL, self._gis, self)
            else:
                rd = {"/rest/": "/admin/", "/SceneServer": ".SceneServer"}
                adminURL = self._str_replace(self._url, rd)
                if adminURL.split("/")[-1].isdigit():
                    adminURL = adminURL.replace(f'/{adminURL.split("/")[-1]}', "")
                self._admin = EnterpriseSceneLayerManager(adminURL, self._gis, self)
        return self._admin

    # ----------------------------------------------------------------------
    def _str_replace(self, mystring, rd):
        """Replaces a value based on a key/value pair where the
        key is the text to replace and the value is the new value.

        The find/replace is case-insensitive.

        """
        import re

        patternDict = {}
        for key, value in rd.items():
            pattern = re.compile(re.escape(key), re.IGNORECASE)
            patternDict[value] = pattern
        for key in patternDict:
            regex_obj = patternDict[key]
            mystring = regex_obj.sub(key, mystring)
        return mystring


###########################################################################
class _SceneLayerFactory(type):
    """
    Factory that generates the Scene Layers

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional :class:`~arcgis.gis.GIS`  object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.PointCloudLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """

    def __call__(cls, url, gis=None):
        lyr = Layer(url=url, gis=gis)
        props = lyr.properties
        if "sublayers" in props:
            return BuildingLayer(url=url, gis=gis)
        elif "layerType" in props:
            lt = props.layerType
        else:
            lt = props.layers[0].layerType
        if str(lt).lower() == "pointcloud":
            return PointCloudLayer(url=url, gis=gis)
        elif str(lt).lower() == "point":
            return Point3DLayer(url=url, gis=gis)
        elif str(lt).lower() == "3dobject":
            return Object3DLayer(url=url, gis=gis)
        elif str(lt).lower() == "building":
            return BuildingLayer(url=url, gis=gis)
        elif str(lt).lower() == "IntegratedMesh".lower():
            return IntegratedMeshLayer(url=url, gis=gis)
        elif str(lt).lower() == "voxel":
            return VoxelLayer(url=url, gis=gis)
        return lyr


###########################################################################
class SceneLayer(Layer, metaclass=_SceneLayerFactory):
    """
    The ``SceneLayer`` class represents a Web scene layer.

    .. note::
        Web scene layers are cached web layers that are optimized for displaying a large amount of 2D and 3D features.

    .. note::
        Web scene layers can be used to represent 3D points, point clouds, 3D objects and
        integrated mesh layers.

    ==================     ====================================================================
    **Parameter**           **Description**
    ------------------     --------------------------------------------------------------------
    url                    Required string, specify the url ending in /SceneServer/
    ------------------     --------------------------------------------------------------------
    gis                    Optional :class:`~arcgis.gis.GIS`  object. If not specified, the active GIS connection is
                           used.
    ==================     ====================================================================

    .. code-block:: python

        # USAGE EXAMPLE 1: Instantiating a SceneLayer object

        from arcgis.mapping import SceneLayer
        s_layer = SceneLayer(url='https://your_portal.com/arcgis/rest/services/service_name/SceneServer/')

        type(s_layer)
        >> arcgis.mapping._types.PointCloudLayer

        print(s_layer.properties.layers[0].name)
        >> 'your layer name'
    """

    def __init__(self, url, gis=None):
        """
        Constructs a SceneLayer given a web scene layer URL
        """
        super(SceneLayer, self).__init__(url, gis)
